# -*- coding: utf-8 -*-

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.



from io import StringIO

from .generic import decide_merge_with_diff
from .decisions import apply_decisions
from ..diffing.notebooks import diff_notebooks
from ..utils import Strategies
from ..prettyprint import (
    pretty_print_notebook_diff,
    pretty_print_merge_decisions,
    pretty_print_notebook,
    PrettyPrintConfig
)

import nbdime.log


# Strategies for handling conflicts
generic_conflict_strategies = (
    "clear",            # Replace value with empty in case of conflict
    "remove",           # Discard value in case of conflict
    "clear-all",        # Discard all values on conflict
    "fail",             # Unexpected: crash and burn in case of conflict (only implemented for leaf nodes)
    "inline-cells",     # Valid for cell only: use markdown cells as diff markers for conflicting inserts/replace
    "inline-source",    # Valid for source only: produce new source with inline diff markers
    "inline-outputs",   # Valid for outputs only: produce new outputs with inline diff markers
    "mergetool",        # Do not modify decision (but prevent processing at deeper path)
    "record-conflict",  # Valid for metadata only: produce new metadata with conflicts recorded for external inspection
    "take-max",         # Take the maximum value in case of conflict
    "union",            # Join values in case of conflict, don't insert new markers (only applies to sequence types)
    "use-base",         # Keep base value in case of conflict
    "use-local",        # Use local value in case of conflict
    "use-remote",       # Use remote value in case of conflict
    )

# Strategies that can be applied to an entire notebook
cli_conflict_strategies = (
    "inline",           # Inline cells or source and outputs, and record metadata conflicts
    "use-base",         # Keep base value in case of conflict
    "use-local",        # Use local value in case of conflict
    "use-remote",       # Use remote value in case of conflict
    #"union",            # Take local value, then remote, in case of conflict
    )

cli_conflict_strategies_input = cli_conflict_strategies

cli_conflict_strategies_output = cli_conflict_strategies + (
    "remove",     # Remove conflicting outputs
    "clear-all",  # Clear all outputs
    )


def notebook_merge_strategies(args):
    strategies = Strategies({
        "/cells/*/id": "remove",
        # These fields should never conflict, that would be an internal error:
        "/nbformat": "fail",
        "/cells/*/cell_type": "fail",
        # Pick highest minor format:
        "/nbformat_minor": "take-max",
        })

    ignore_transients = args.ignore_transients if args else True
    if ignore_transients:
        strategies.transients = [
            "/cells/*/execution_count",
            #"/cells/*/outputs",
            "/cells/*/outputs/*/execution_count",
            "/cells/*/metadata/collapsed",
            "/cells/*/metadata/autoscroll",
            "/cells/*/metadata/scrolled",
        ]
        strategies.update({
            "/cells/*/execution_count": "clear",
            "/cells/*/outputs/*/execution_count": "clear",
        })

    # Get args, default to inline for cli tool, intended to produce
    # an editable notebook that can be manually edited
    merge_strategy = args.merge_strategy if args else "inline"
    input_strategy = args.input_strategy if args else None
    output_strategy = args.output_strategy if args else None

    # Default to merge_strategy
    input_strategy = input_strategy or merge_strategy
    output_strategy = output_strategy or merge_strategy
    metadata_strategy = merge_strategy if merge_strategy != "union" else None

    # Set root strategy
    if merge_strategy == 'inline':
        strategies['/cells'] = "inline-cells"
    elif merge_strategy == 'union':
        strategies['/cells'] = merge_strategy
    else:
        strategies["/"] = merge_strategy

    # Translate 'inline' to specific strategies for different fields
    if input_strategy == 'inline':
        source_strategy = "inline-source"
        attachments_strategy = "inline-attachments"
    else:
        source_strategy = input_strategy
        attachments_strategy = input_strategy

    if output_strategy == 'inline':
        outputs_strategy = "inline-outputs"
    else:
        outputs_strategy = output_strategy

    if metadata_strategy == "inline":
        metadata_strategy = "record-conflict"

    # Set strategies on the main fields
    strategies.update({
        "/metadata": metadata_strategy,
        "/cells/*/metadata": metadata_strategy,
        "/cells/*/outputs/*/metadata": metadata_strategy,
        "/cells/*/source": source_strategy,
        "/cells/*/attachments": attachments_strategy,
        "/cells/*/outputs": outputs_strategy
    })

    return strategies


def decide_notebook_merge(base, local, remote, args=None):
    # Build merge strategies for each document path from arguments
    strategies = notebook_merge_strategies(args)

    # Compute notebook specific diffs
    local_diffs = diff_notebooks(base, local)
    remote_diffs = diff_notebooks(base, remote)

    # Debug outputs
    if args and args.log_level == "DEBUG":
        # log pretty-print config object:
        config = PrettyPrintConfig()

        nbdime.log.debug("In merge, base-local diff:")
        config.out = StringIO()
        pretty_print_notebook_diff("<base>", "<local>", base, local_diffs, config)
        nbdime.log.debug(config.out.getvalue())

        nbdime.log.debug("In merge, base-remote diff:")
        config.out = StringIO()
        pretty_print_notebook_diff("<base>", "<remote>", base, remote_diffs, config)
        nbdime.log.debug(config.out.getvalue())

    # Execute a generic merge operation
    decisions = decide_merge_with_diff(
        base, local, remote,
        local_diffs, remote_diffs,
        strategies)

    # Debug outputs
    if args and args.log_level == "DEBUG":
        nbdime.log.debug("In merge, decisions:")
        config.out = StringIO()
        pretty_print_merge_decisions(base, decisions, config)
        nbdime.log.debug(config.out.getvalue())

    return decisions


def merge_notebooks(base, local, remote, args=None):
    """Merge changes introduced by notebooks local and remote from a shared ancestor base.

    Return new (partially) merged notebook and unapplied diffs from the local and remote side.
    """
    if args and args.log_level == "DEBUG":
        # log pretty-print config object:
        config = PrettyPrintConfig()
        for (name, nb) in [("base", base), ("local", local), ("remote", remote)]:
            nbdime.log.debug("In merge, input %s notebook:", name)
            config.out = StringIO()
            pretty_print_notebook(nb, config)
            nbdime.log.debug(config.out.getvalue())

    decisions = decide_notebook_merge(base, local, remote, args)

    merged = apply_decisions(base, decisions)

    if args and args.log_level == "DEBUG":
        nbdime.log.debug("In merge, merged notebook:")
        config.out = StringIO()
        pretty_print_notebook(merged, config)
        nbdime.log.debug(config.out.getvalue())
        nbdime.log.debug("End merge")

    return merged, decisions



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import copy
import nbformat
from nbformat import NotebookNode

from .diff_format import DiffOp, NBDiffFormatError
from .diff_utils import flatten_list_of_string_diff



__all__ = ["patch", "patch_notebook"]


def patch_list(obj, diff):
    # The patched sequence to build and return
    newobj = []
    # Index into obj, the next item to take unless diff says otherwise
    take = 0
    for e in diff:
        op = e.op
        index = e.key
        assert isinstance(index, int), 'list key must be integer'

        # Take values from obj not mentioned in diff, up to not including index
        newobj.extend(copy.deepcopy(value) for value in obj[take:index])

        if op == DiffOp.ADDRANGE:
            # Extend with new values directly
            newobj.extend(e.valuelist)
            skip = 0
        elif op == DiffOp.REMOVERANGE:
            # Delete a number of values by skipping
            skip = e.length
        elif op == DiffOp.PATCH:
            newobj.append(patch(obj[index], e.diff))
            skip = 1
        # Note that the operations ADD, REMOVE, REPLACE are not produced by the
        # diff algorithm anymore, keeping these cases just in case we want them back:
        elif op == DiffOp.ADD:
            # Append new value directly
            newobj.append(e.value)
            skip = 0
        elif op == DiffOp.REMOVE:
            # Delete values obj[index] by incrementing take to skip
            skip = 1
        elif op == DiffOp.REPLACE:
            # Add replacement value and skip old
            newobj.append(e.value)
            skip = 1
        else:
            raise NBDiffFormatError("Invalid op {}.".format(op))

        # Skip the specified number of elements, but never decrement take.
        # Note that take can pass index in diffs with repeated +/- on the
        # same index, i.e. [op_remove(index), op_add(index, value)]
        take = max(take, index + skip)

    # Take values at end not mentioned in diff
    newobj.extend(copy.deepcopy(value) for value in obj[take:len(obj)])

    return newobj


def patch_string(obj, diff):
    "Patch a multiline string, assuming diff is line based."
    # This can possibly be optimized for str if wanted, but
    # waiting until patch_list has been tested and debugged better

    # Flatten line-based diff to character based first!
    diff = flatten_list_of_string_diff(obj, diff)
    return "".join(patch_list(list(obj), diff))


def patch_singleline_string(obj, diff):
    "Patch a singleline string, assuming diff is character based."
    # This can possibly be optimized for str if wanted, but
    # waiting until patch_list has been tested and debugged better
    return "".join(patch_list(list(obj), diff))


def patch_dict(obj, diff):
    newobj = {}
    deleted_keys = set()

    for e in diff:
        op = e.op
        key = e.key
        assert isinstance(key, str), 'dict key must be string'
        assert key not in newobj, 'multiple diff entries target same key: %r' % key

        if op == DiffOp.ADD:
            assert key not in obj, 'patch add value not found for key: %r' % key
            newobj[key] = e.value
        elif op == DiffOp.REMOVE:
            deleted_keys.add(key)
        elif op == DiffOp.REPLACE:
            assert key not in deleted_keys, 'cannot replace deleted key: %r' % key
            newobj[key] = e.value
        elif op == DiffOp.PATCH:
            assert key not in deleted_keys, 'cannot patch deleted key: %r' % key
            newobj[key] = patch(obj[key], e.diff)
        else:
            raise NBDiffFormatError("Invalid op {}.".format(op))

    # Take items not mentioned in diff
    for key in obj:
        if key not in deleted_keys and key not in newobj:
            newobj[key] = copy.deepcopy(obj[key])

    return NotebookNode(newobj)


def patch(obj, diff):
    """Produce a patched version of obj with given hierarchical diff.

    A valid input object can be any dict or list of leaf values,
    or arbitrarily nested dict or list of valid input objects.

    Dicts are required to have string keys.

    Leaf values are any non-dict, non-list objects as far as patch
    is concerned, although the intentional use of this library
    is that values are json-serializable.
    """
    if isinstance(obj, dict):
        return patch_dict(obj, diff)
    elif isinstance(obj, list):
        return patch_list(obj, diff)
    elif isinstance(obj, str):
        return patch_string(obj, diff)
    else:
        raise ValueError("Invalid object type to patch: {}".format(type(obj).__name__))


def patch_notebook(nb, diff):
    return nbformat.from_dict(patch(nb, diff))



# coding: utf-8

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

diff_ignorables = ('sources', 'outputs', 'attachments', 'metadata', 'id', 'details')




import os
import shutil
import tempfile

from nbdime.utils import (
    strings_to_lists, revert_strings_to_lists, is_in_repo,
    locate_gitattributes
)


def test_string_to_lists():
    obj = {"c": [{"s": "ting\ntang", "o": [{"ot": "stream"}]}]}
    obj2 = strings_to_lists(obj)
    obj3 = revert_strings_to_lists(obj2)
    assert obj3 == obj


def test_is_repo():
    try:
        tmpdir = tempfile.mkdtemp(prefix='nbdime-test')
        subdir = tempfile.mkdtemp(dir=tmpdir)
        subfile = tempfile.NamedTemporaryFile(dir=tmpdir)
        subsubfile = tempfile.NamedTemporaryFile(dir=subdir)
        with subfile, subsubfile:
            assert is_in_repo(tmpdir) is False
            assert is_in_repo(subdir) is False
            assert is_in_repo(subfile.name) is False
            assert is_in_repo(subsubfile.name) is False
            os.makedirs(os.path.join(subdir, '.git'))
            assert is_in_repo(tmpdir) is False
            assert is_in_repo(subdir) is True
            assert is_in_repo(subfile.name) is False
            assert is_in_repo(subsubfile.name) is True
            os.makedirs(os.path.join(tmpdir, '.git'))
            assert is_in_repo(tmpdir) is True
            assert is_in_repo(subdir) is True
            assert is_in_repo(subfile.name) is True
            assert is_in_repo(subsubfile.name) is True

    finally:
        shutil.rmtree(tmpdir)


def test_locate_gitattributes_local(git_repo):
    gitattr = locate_gitattributes(scope=None)
    assert gitattr is not None


def test_locate_gitattributes_global(needs_git):
    gitattr = locate_gitattributes(scope='global')
    assert gitattr is not None


def test_locate_gitattributes_system(needs_git):
    gitattr = locate_gitattributes(scope='system')
    assert gitattr is not None



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import copy
import nbformat

import nbdime.log
from ..diff_format import (
    DiffOp, op_removerange, op_remove, op_patch, op_replace)
from ..patching import patch
from ..utils import (
    r_is_int, star_path, join_path, is_prefix_array, find_shared_prefix)

def as_list(x):
    if x is None:
        return x
    if isinstance(x, tuple):
        return list(x)
    if not isinstance(x, list):
        return [x]
    return x

class MergeDecision(dict):
    """For internal usage in nbdime library.

    Minimal class providing attribute access to merge decision keys.

    Tip: If performance dictates, we can easily replace this
    with a namedtuple during processing of diffs and convert
    to dicts before any json conversions.
    """

    def __getattr__(self, name):
        if name.startswith("__") and name.endswith("__"):
            return self.__getattribute__(name)
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def local_path(self):
        level = self.get('_level', 0)
        return (self.common_path or ())[level:]


class MergeDecisionBuilder(object):
    """A helper class for building a series of decisions to describe a merge.
    """
    def __init__(self):
        self.decisions = []

    def __bool__(self):
        return bool(self.decisions)

    def __nonzero__(self):
        return bool(self.decisions)

    def __len__(self):
        return len(self.decisions)

    def __iter__(self):
        return iter(self.decisions)

    def validated(self, base):
        """Returns decisions in state ready for application.

        Most importantly, this sorts the decisions on the path, so that it is
        in accordance with the specs.
        """
        # Remove fields 'strategy' used for internal decision making but not part of spec
        for d in self.decisions:
            if "strategy" in d:
                del d["strategy"]
        return sorted(self.decisions, key=_sort_key, reverse=True)

    def extend(self, decisions):
        if isinstance(decisions, MergeDecisionBuilder):
            decisions = decisions.decisions
        self.decisions.extend(decisions)

    def get_conflicted(self):
        return [d for d in self.decisions if d.conflict]

    def has_conflicted(self):
        return any(d.conflict for d in self.decisions)

    def add_decision(self, path, action, local_diff, remote_diff,
                     conflict=False, strategy=None, **kwargs):
        """Add a decision to the builder with the specified properties.

        Ensures data types and paths are as they should be, before creating a
        MergeDecision and adding it to its internal store.
        """
        # Ensure path is immutable
        if isinstance(path, list):
            path = tuple(path)
        else:
            assert isinstance(path, tuple), 'decision paths should be tuples'
        custom_diff = kwargs.pop("custom_diff", None)
        # Ensure diffs are lists or None
        local_diff = as_list(local_diff)
        remote_diff = as_list(remote_diff)
        custom_diff = as_list(custom_diff)
        # Ensure paths are pushed out as far in tree as possible
        path, (local_diff, remote_diff, custom_diff) = \
            ensure_common_path(path, [local_diff, remote_diff, custom_diff])
        if custom_diff is not None:
            kwargs["custom_diff"] = custom_diff
        # Store strategy field only if given
        if strategy is not None:
            kwargs["strategy"] = strategy

        # Finally store decision
        self.decisions.append(MergeDecision(
            common_path=path,
            conflict=conflict,
            action=action,
            local_diff=local_diff,
            remote_diff=remote_diff,
            **kwargs
            ))

    def base(self, path, local_diff, remote_diff, conflict=False, strategy=None):
        self.add_decision(
            path=path,
            action="base",
            conflict=conflict,
            local_diff=local_diff,
            remote_diff=remote_diff,
            strategy=strategy
            )

    def onesided(self, path, local_diff, remote_diff, conflict=False):
        assert local_diff or remote_diff, 'one diff needed in onesided merge decisions'
        assert not (local_diff and remote_diff), (
            'one diff should be empty in onesided merge decisions')
        if local_diff:
            action = "local"
        elif remote_diff:
            action = "remote"
        self.add_decision(
            path=path,
            action=action,
            local_diff=local_diff,
            remote_diff=remote_diff,
            conflict=conflict,
            )

    def local_then_remote(self, path, local_diff, remote_diff, conflict=False, strategy=None):
        assert local_diff and remote_diff, 'should have two diffs for sequential merge decisions'
        action = "local_then_remote"
        self.add_decision(
            path=path,
            conflict=conflict,
            action=action,
            local_diff=local_diff,
            remote_diff=remote_diff,
            strategy=strategy
            )

    def remote_then_local(self, path, local_diff, remote_diff, conflict=False, strategy=None):
        assert local_diff and remote_diff, 'should have two diffs for sequential merge decisions'
        action = "remote_then_local"
        self.add_decision(
            path=path,
            conflict=conflict,
            action=action,
            local_diff=local_diff,
            remote_diff=remote_diff,
            strategy=strategy
            )

    def agreement(self, path, local_diff, remote_diff, conflict=False):
        assert local_diff and remote_diff, 'should have two diffs for agreed merge decisions'
        assert local_diff == remote_diff, 'should have identical diffs for agreed merged decisions'
        self.add_decision(
            path=path,
            action="either",
            local_diff=local_diff,
            remote_diff=remote_diff,
            conflict=conflict,
            )

    def tryresolve(self, path, local_diff, remote_diff, strategy):
        """Try to resolve conflict with given strategy.

        If successful, registers a decision and returns the action used.
        If failing, does not register a decision and returns None.

        Valid strategies here are:
            use-local, use-remote, use-base, clear, take-max
        """
        if not strategy:
            return None

        assert local_diff and remote_diff, 'onesided merges should not be conflicted'
        assert local_diff != remote_diff, 'agreed merges should not be conflicted'

        # Allow strategies to defuse situation first
        action = None
        if strategy:
            # Applying strategies here works well for leaf nodes at least
            if strategy == "use-local":
                action = "local"
            elif strategy == "use-remote":
                action = "remote"
            elif strategy == "use-base":
                action = "base"
            elif strategy == "union":
                action = "local_then_remote"
            elif strategy == "clear":
                action = "clear"
            elif strategy == "take-max":
                action = "take_max"
            elif strategy == "fail":
                msg = "Unexpected conflict on {}.".format(path)
                nbdime.log.error(msg)
                raise RuntimeError(msg)
            else:
                msg = "Unhandled conflict strategy {} on {}".format(
                    strategy, join_path(path))
                nbdime.log.warning(msg)

        if action is not None:
            self.add_decision(
                path=path,
                conflict=False,
                action=action,
                local_diff=local_diff,
                remote_diff=remote_diff,
                strategy=strategy
                )
        return action

    def conflict(self, path, local_diff, remote_diff, strategy=None):
        """Register a potential conflict.

        If strategy is given, tries to resolve conflict first with tryresolve.

        Complex strategies not handled by tryresolve need to be handled at
        an earlier stage and will result in a regular conflict here.
        """
        assert local_diff and remote_diff, 'onesided merges should not be conflicted'
        assert local_diff != remote_diff, 'agreed merges should not be conflicted'

        # Try to defuse situation with given strategy
        action = self.tryresolve(path, local_diff, remote_diff, strategy)

        # If none of them applied, use base and mark as conflict
        if action is None:
            # If a strategy was provided but failed to apply, mark as conflict.
            # NB! Not passing strategy argument on to decision because it hasn't been applied.
            self.add_decision(
                path=path,
                conflict=True,
                action="base",
                local_diff=local_diff,
                remote_diff=remote_diff
                )

    def local(self, path, local_diff, remote_diff, conflict=False, strategy=None):
        assert local_diff, 'needs non-empty local diff'
        action = "local"
        self.add_decision(
            path=path,
            action=action,
            conflict=conflict,
            local_diff=local_diff,
            remote_diff=remote_diff,
            strategy=strategy
            )

    def remote(self, path, local_diff, remote_diff, conflict=False, strategy=None):
        assert remote_diff, 'needs non-empty remote diff'
        action = "remote"
        self.add_decision(
            path=path,
            action=action,
            conflict=conflict,
            local_diff=local_diff,
            remote_diff=remote_diff,
            strategy=strategy
            )

    def custom(self, path, local_diff, remote_diff, custom_diff, conflict=False, strategy=None):
        action = "custom"
        self.add_decision(
            path=path,
            conflict=conflict,
            action=action,
            local_diff=local_diff,
            remote_diff=remote_diff,
            custom_diff=custom_diff,
            strategy=strategy
            )

    def similar_insert(self, path, local_diff, remote_diff, insert_diff, strategy=None):
        """Same as `conflict`, but with marker `similar_insert``"""
        assert local_diff and remote_diff, 'onesided merges should not be conflicted'
        assert local_diff != remote_diff, 'agreed merges should not be conflicted'

        # Try to defuse situation with given strategy
        action = self.tryresolve(path, local_diff, remote_diff, strategy)

        # If none of them applied, use base and mark as conflict
        if action is None:
            # If a strategy was provided but failed to apply, mark as conflict.
            # NB! Not passing strategy argument on to decision because it hasn't been applied.
            self.add_decision(
                path=path,
                conflict=True,
                action="base",
                local_diff=local_diff,
                remote_diff=remote_diff,
                similar_insert=insert_diff
                )


def ensure_common_path(path, diffs):
    """Resolves common paths in a list of diffs.

    If a local and a remote diff both patch a key "a", this will return the
    common path ("a",), and the inner diffs of the patch operations. Works
    recursively, so a common chain of patches will be resolved as well.
    """
    assert isinstance(path, (tuple, list)), 'incorrect path type'
    popped = _pop_path(diffs)
    while popped:
        path = path + (popped["key"],)
        diffs = popped["diffs"]
        popped = _pop_path(diffs)
    return path, diffs


def _pop_path(diffs):
    """Pops of a common path from patch ops sharing a common key.

    Checks whether all diffs are single patch operations sharing the same key,
    or alternatively empty diffs. If so, it returns the shared path/key as well
    as the inner diffs of the patch operations (in the same order as the passed
    diffs).
    """
    key = None
    popped_diffs = []
    for d in diffs:
        # Empty diffs can be skipped
        if d is None or len(d) == 0:
            popped_diffs.append(None)
            continue
        # Check that we have only one op, which is a patch op
        if len(d) != 1 or d[0].op != DiffOp.PATCH:
            return
        # Ensure all present diffs have the same key
        if key is None:
            key = d[0].key
        elif key != d[0].key:
            return
        # Ensure the sub diffs of all ops are suitable as outer layer
        # if len(d[0].diff) > 1:
        #    return
        popped_diffs.append(d[0].diff)
    if key is None:
        return
    return {'key': key, 'diffs': popped_diffs}


def push_path(path, diffs):
    """Wraps the diffs in patch operations matching path.
    """
    for key in reversed(path):
        diffs = [op_patch(key, diffs)]
    return diffs


def pop_patch_decision(decision):
    """Create a new decision one level lower in diff tree.

    Checks whether all diffs are single patch operations sharing the same key,
    or alternatively empty diffs. A new decision is then created at that key.

    Returns the new decision.

    Returns None if a decision can not be created at the lower level.
    """
    diffs = [decision.local_diff, decision.remote_diff]
    if decision.action == "custom":
        diffs.append(decision.custom_diff)
    popped = _pop_path(diffs)
    if popped is None:
        return None
    ret = MergeDecision(
        common_path=decision.common_path + (popped["key"],),
        local_diff=popped["diffs"][0],
        remote_diff=popped["diffs"][1],
        action=decision.action,
        conflict=decision.conflict)
    if decision.action == "custom":
        ret.custom_diff = popped["diffs"][2]
    return ret


def pop_all_patch_decisions(decision):
    """Create a new decision at the furthest level in the diff tree.

    Calls `pop_patch_decision` recursively until it is at the lowest possible
    decision level.

    If the decision is already at the lowest level, it returns the original
    decision.
    """
    popped = pop_patch_decision(decision)
    while popped is not None:
        decision = popped
        popped = pop_patch_decision(decision)
    return decision


def push_patch_decision(decision, prefix):
    """Move a path prefix in a merge decision from `common_path` to the diffs.

    This is done by wrapping the diffs in nested patch ops.
    """
    dec = copy.copy(decision)
    # We need to start with inner most key to nest correctly, so reverse:
    for key in reversed(prefix):
        if len(dec.common_path) == 0:
            raise ValueError(
                "Cannot remove key from empty decision path: %r, %r" %
                (key, dec))
        assert dec.common_path[-1] == key, "Key %r not at end of %r" % (
            key, dec.common_path)
        dec.common_path = dec.common_path[:-1]  # pop key
        dec.local_diff = [op_patch(key, dec.local_diff)] if dec.local_diff else []
        dec.remote_diff = [op_patch(key, dec.remote_diff)] if dec.remote_diff else []
        if dec.action == "custom":
            dec.custom_diff = [op_patch(key, dec.custom_diff)] if dec.custom_diff else []
    return dec


def _sort_key(k):
    """Sort key for common paths. Ensures the correct order for processing,
    without having to care about offsetting indices.

    Heavily inspired by the natsort package:

    Copyright (c) 2012-2016 Seth M. Morton

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to
    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """
    ret = []
    for s in k.common_path:
        if not isinstance(s, (int, str)):
            s = s.decode("utf8")
        if isinstance(s, str) and r_is_int.match(s):
            s = int(s)
        if isinstance(s, int):
            ret.append(('', -s))
        else:
            ret.append((s,))
    return ret


def split_string_path(base, path):
    """Prevent paths from pointing to specific string lines.

    Since strings are diffed as a sequence of lines without actually splitting
    the string in base, any path to a specific line will fail to resolve.
    This checks if path points to a specific line in a string, and splits off
    the final key of the path (the line number).

    Returns a tuple with the path without any line reference in the first
    position, and any line key in the second position.
    """
    for i in range(len(path)):
        if isinstance(base, str):
            return path[:i], path[i:]
        base = base[path[i]]
    return path, ()


def make_cleared_value(value):
    "Make a new 'cleared' value of the right type."
    if isinstance(value, list):
        # Clearing e.g. an outputs list means setting it to an empty list
        return []
    elif isinstance(value, dict):
        # Clearing e.g. a metadata dict means setting it to an empty dict
        return {}
    elif isinstance(value, str):
        # Clearing e.g. a source string means setting it to an empty string
        return ""
    else:
        # Clearing anything else (atomic values) means setting it to None
        return None


def filter_decisions(pattern, decisions, exact=False):
    ret = []
    cutoff = len(pattern)
    for i, md in enumerate(decisions):
        path = md.common_path[:]
        pop = _pop_path((md.local_diff, md.remote_diff, md.get('custom_diff')))
        if pop:
            path = path + (pop["key"],)
        starred_path = star_path(path)
        if (exact and starred_path == pattern or
                starred_path[:cutoff] == pattern):
            ret.append(i)
    return ret


# =============================================================================
#
# Code for applying decisions:
#
# =============================================================================

# # Strategies for handling conflicts
# generic_conflict_strategies = (
#     # on /cells/*/source, string in a dict, let decision be placed on the cell dict!
#     "inline-source",    # Valid for source only: produce new source with inline diff markers
#     # on /cells/*/outputs, list in a dict, let decision be placed on the cell dict!
#     "clear-all",        # Discard all values on conflict
#     # on /cells/*/outputs, list in a dict, let decision be placed on the cell dict!
#     "inline-outputs",   # Valid for outputs only: produce new outputs with inline diff markers
#     # on /.../metadata, add to a dict, let decision be placed on the cell dict!
#     "record-conflict",  # Valid for metadata only: produce new metadata with conflicts recorded for external inspection
#     )

def resolve_action(base, decision):
    a = decision.action

    if a == "base":
        return []   # no-op

    elif a in ("local", "either"):
        return copy.copy(decision.local_diff)

    elif a == "remote":
        return copy.copy(decision.remote_diff)

    elif a == "custom":
        return copy.copy(decision.custom_diff)

    elif a == "local_then_remote":
        return decision.local_diff + decision.remote_diff

    elif a == "remote_then_local":
        return decision.remote_diff + decision.local_diff

    elif a in ("clear", "remove"):
        key, = set(d.key for d in decision.local_diff + decision.remote_diff)
        if a == 'clear':
            return [op_replace(key, make_cleared_value(base[key]))]
        elif isinstance(base, (list, str)):
            return [op_removerange(key, 1)]
        else:
            return [op_remove(key)]

    elif a == "clear_all":
        if isinstance(base, dict):
            # Ideally we would do a op_replace on the parent, but this is not
            # easily combined with this method, so simply remove all keys
            return [op_remove(key) for key in base.keys()]
        elif isinstance(base, (list, str)):
            return [op_removerange(0, len(base))]

    elif a == "take_max":
        key, = set(d.key for d in decision.local_diff + decision.remote_diff)
        #assert len(decision.local_diff) == 1 == len(decision.remote_diff)
        bval = base[key]
        lval = decision.local_diff[0].value if decision.local_diff else bval
        rval = decision.remote_diff[0].value if decision.remote_diff else bval
        mval = max(bval, lval, rval)
        if bval == mval:
            return []
        else:
            return [op_replace(key, mval)]

    else:
        raise NotImplementedError("The action \"%s\" is not defined" % a)


def apply_decisions(base, decisions):
    """Apply a list of merge decisions to base.
    """
    from .strategies import combine_patches

    merged = copy.deepcopy(base)
    prev_path = None
    parent = None
    last_key = None
    resolved = None
    diffs = []
    # clear_all actions should override other decisions on same obj, so
    # we need to track it
    clear_all_flag = False
    for md in decisions:
        path, line = split_string_path(merged, md.common_path)
        # We patch all decisions with the same path in one op
        if path == prev_path:
            # Same path as previous, collect entry
            if clear_all_flag:
                # Another entry will clear the parent, all other decisions
                # should be dropped
                pass
            else:
                if md.action == "clear_all":
                    clear_all_flag = True
                    # Clear any existing decisions!
                    diffs = []
                ad = resolve_action(resolved, md)
                if line:
                    ad = push_path(line, ad)
                diffs = combine_patches(diffs + ad)

        else:
            # Different path, start a new collection
            if prev_path is not None:
                # First, apply previous diffs
                if parent is None:
                    # Operations on root create new merged object
                    merged = patch(resolved, diffs)
                else:
                    # If not, overwrite entry in parent (which is an entry in
                    # merged). This is ok, as no paths should point to
                    # subobjects of the patched object
                    parent[last_key] = patch(resolved, diffs)

            prev_path = path
            # Resolve path in base and output
            resolved = merged
            parent = None
            last_key = None
            for key in path:
                parent = resolved
                resolved = resolved[key]   # Should raise if key missing
                last_key = key
            diffs = resolve_action(resolved, md)
            if line:
                diffs = push_path(line, diffs)
            clear_all_flag = md.action == "clear_all"
    # Apply the last collection of diffs, if present (same as above)
    if prev_path is not None:
        if parent is None:
            merged = patch(resolved, diffs)
        else:
            parent[last_key] = patch(resolved, diffs)

    merged = nbformat.from_dict(merged)
    return merged


def _merge_tree(tree, sorted_paths):
    """
    Merge a tree of diffs at varying path levels to one diff at their shared root

    Relies on the format specification about decision ordering to help
    simplify the process (deeper paths should come before its parent paths).
    This is realized by the `sorted_paths` argument.
    """
    trunk = []
    root = None
    for i in range(len(sorted_paths)):
        path = tree[sorted_paths[i]]['path']
        subdiffs = tree[sorted_paths[i]]['diff']
        trunk = trunk + subdiffs

        if i == len(sorted_paths) - 1:
            nextPath = root
        else:
            nextPath = tree[sorted_paths[i + 1]]['path']

        # First, check if path is subpath of nextPath:
        if is_prefix_array(nextPath, path):
            # We can simply promote existing diffs to next path
            if nextPath is not None:
                trunk = push_path(path[len(nextPath):], trunk)
                root = nextPath
        else:
            # We have started on a new trunk
            # Collect branches on the new trunk, and merge the trunks
            newTrunk = _merge_tree(tree, sorted_paths[i + 1:])
            nextPath = tree[sorted_paths[len(sorted_paths) - 1]]['path']
            prefix = find_shared_prefix(path, nextPath)
            pl = len(prefix) if prefix is not None else 0
            trunk = push_path(path[pl:], trunk) + push_path(nextPath[pl:], newTrunk)
            break   # Recursion will exhaust sorted_paths
    return trunk


def build_diffs(base, decisions, which):
    """
    Builds a diff for direct application on base. The `which` argument either
    selects the 'local', 'remote' or 'merged' diffs.
    """
    # First, we translate the decisions to a dict tree of the format
    # path (string): {diff: diff entries, path: path (tuple)}:
    tree = {}
    # The paths of the tree in sorted order (order of first appearance in decisions).
    # Will be sorted in order deeper paths first for decisions that share a common
    # path prefix.
    sorted_paths = []
    assert which in ('local', 'remote', 'merged'), 'invalid argument %r' % which
    local = which == 'local'
    merged = which == 'merged'

    for md in decisions:
        # The path might include string line number, split those off:
        path, line = split_string_path(base, md.local_path())
        # Get the diff for the current decision:
        if merged:
            subdiffs = resolve_action(base[path], md)
        else:
            if local:
                subdiffs = md.local_diff
            else:
                subdiffs = md.remote_diff
            if subdiffs is None:
                continue

        str_path = join_path(path)
        if str_path in tree:
            # Existing tree entry, simply add diffs to it
            if line:
                match_diff = [d for d in tree[str_path]['diff'] if d.key == line[0]]
                if match_diff:
                    assert len(match_diff) == 1, 'multiple mathing diffs found in tree'
                    assert match_diff[0].diff, 'empty diff found in tree'
                    match_diff[0].diff.extend(subdiffs)
                else:
                    subdiffs = push_path(line, subdiffs)
                    assert len(subdiffs) == 1
                    tree[str_path]['diff'].extend(subdiffs)
            else:
                tree[str_path]['diff'].extend(subdiffs)
        else:
            # Make new entry in tree
            if line:
                subdiffs = push_path(line, subdiffs)
            tree[str_path] = {'diff': subdiffs, 'path': path}
            sorted_paths.append(str_path)

    if len(tree) == 0:
        return None

    if '/' not in tree:
        tree['/'] = {'diff': [], 'path': []}
        sorted_paths.append('/')

    # Tree is constructed, now join all branches at diverging points (joints)
    return _merge_tree(tree, sorted_paths)



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.



from nbdime import merge_notebooks

from nbdime.nbmergeapp import _build_arg_parser

# FIXME: Extend tests to more merge situations!


# Tests here assume default autoresolve behaviour at time of writing,
# this is likely to change and it's ok to update the tests to reflect
# new behaviour as needed!


# Setup default args for merge app
builder = _build_arg_parser()


def test_autoresolve_notebook_ec():
    # Tests here assume default autoresolve behaviour at time of writing,
    # this may change and tests should then be updated
    args = None

    source = "def foo(x, y):\n    return x**y"
    base = {"cells": [{
        "source": source, "execution_count": 1, "cell_type": "code",
        "outputs": None}]}
    local = {"cells": [{
        "source": source, "execution_count": 2, "cell_type": "code",
        "outputs": None}]}
    remote = {"cells": [{
        "source": source, "execution_count": 3, "cell_type": "code",
        "outputs": None}]}
    merged, decisions = merge_notebooks(base, local, remote, args)
    assert merged == {"cells": [{
        "source": source, "execution_count": None, "outputs": None,
        "cell_type": "code"}]}
    assert not any(d.conflict for d in decisions)


def test_autoresolve_notebook_no_ignore():
    args = builder.parse_args(["", "", ""])
    args.ignore_transients = False

    source = "def foo(x, y):\n    return x**y"
    base = {"cells": [{
        "source": source, "execution_count": 1, "cell_type": "code",
        "outputs": None}]}
    local = {"cells": [{
        "source": source, "execution_count": 2, "cell_type": "code",
        "outputs": None}]}
    remote = {"cells": [{
        "source": source, "execution_count": 3, "cell_type": "code",
        "outputs": None}]}
    merged, decisions = merge_notebooks(base, local, remote, args)
    assert merged == {"cells": [{
        "source": source, "execution_count": 1, "outputs": None,
        "cell_type": "code"}]}
    assert decisions[0].conflict is True
    assert len(decisions) == 1


def test_autoresolve_notebook_no_ignore_fallback():
    args = builder.parse_args(["", "", ""])
    args.ignore_transients = False
    args.merge_strategy = 'use-remote'

    source = "def foo(x, y):\n    return x**y"
    base = {"cells": [{
        "source": source, "execution_count": 1, "cell_type": "code",
        "outputs": None}]}
    local = {"cells": [{
        "source": source, "execution_count": 2, "cell_type": "code",
        "outputs": None}]}
    remote = {"cells": [{
        "source": source, "execution_count": 3, "cell_type": "code",
        "outputs": None}]}
    merged, decisions = merge_notebooks(base, local, remote, args)
    assert merged == {"cells": [{
        "source": source, "execution_count": 3, "outputs": None,
        "cell_type": "code"}]}
    assert not any(d.conflict for d in decisions)


def test_autoresolve_notebook_ignore_fallback():
    args = builder.parse_args(["", "", ""])
    args.ignore_transients = True
    args.merge_strategy = 'use-remote'

    source = "def foo(x, y):\n    return x**y"
    base = {"cells": [{
        "source": source, "execution_count": 1, "cell_type": "code",
        "outputs": None}]}
    local = {"cells": [{
        "source": source, "execution_count": 2, "cell_type": "code",
        "outputs": None}]}
    remote = {"cells": [{
        "source": source, "execution_count": 3, "cell_type": "code",
        "outputs": None}]}
    merged, decisions = merge_notebooks(base, local, remote, args)
    assert merged == {"cells": [{
        "source": source, "execution_count": None, "outputs": None,
        "cell_type": "code"}]}
    assert not any(d.conflict for d in decisions)


def test_autoresolve_inline_source_conflict(db, needs_git):
    nbb = db["inline-conflict--1"]
    nbl = db["inline-conflict--2"]
    nbr = db["inline-conflict--3"]

    args = builder.parse_args(["", "", ""])
    args.merge_strategy = 'inline'
    merged, decisions = merge_notebooks(nbb, nbl, nbr, args)

    # Has conflicts
    assert any(d.conflict for d in decisions)

    source = merged.cells[0].source

    expected = """\
x = 1
<<<<<<< local
y = 3
z = 4
print(x * y / z)
=======
q = 3.1
print(x + q)
>>>>>>> remote"""

    assert source == expected



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from ..diff_format import DiffOp, SequenceDiffBuilder


def __unused__get_diff_range(diffs, i):
    "Returns diff entry and range j..k which this diff affects, i.e. base[j:k] is affected."
    e = diffs[i]
    j = e.key
    if e.op == DiffOp.PATCH:
        k = j + 1
    elif e.op == DiffOp.ADDRANGE:
        k = j
    elif e.op == DiffOp.REMOVERANGE:
        k = j + e.length
    else:
        raise ValueError("Unexpected diff op {}".format(e.op))
    return e, j, k


def get_section_boundaries(diffs):
    boundaries = set()
    for e in diffs:
        j = e.key
        boundaries.add(j)
        if e.op == DiffOp.ADDRANGE:
            pass
        elif e.op == DiffOp.REMOVERANGE:
            k = j + e.length
            boundaries.add(k)
        elif e.op == DiffOp.PATCH:
            k = j + 1
            boundaries.add(k)
    return boundaries


def split_diffs_on_boundaries(diffs, boundaries):
    newdiffs = SequenceDiffBuilder()
    assert isinstance(boundaries, list), 'boundaries argument should be a list'

    # Next relevant boundary index
    b = 0

    for e in diffs:
        if e.op in (DiffOp.ADDRANGE, DiffOp.PATCH):
            # Nothing to split
            newdiffs.append(e)
        elif e.op == DiffOp.REMOVERANGE:
            # Skip boundaries smaller than key
            while boundaries[b] < e.key:
                b += 1

            # key should be included in the boundaries
            assert boundaries[b] == e.key, 'key not found in boundaries'

            # Add diff entries for each interval between boundaries up to k
            while b < len(boundaries)-1 and boundaries[b + 1] <= e.key + e.length:
                newdiffs.removerange(boundaries[b], boundaries[b + 1] - boundaries[b])
                b += 1
        else:
            raise ValueError("Unhandled diff entry op {}.".format(e.op))

    return newdiffs.validated()


def make_chunks(boundaries, diffs):
    """Make list of chunks on the form (j, k, diffs0, diffs1, ..., diffsN),
    where `j` and `k` are line numbers in the base, and the `diffsX`
    entries are subsets from `diffs` that are part of the chunk.

    Because the diff entries have been split on the union of
    begin/end boundaries of all diff entries, the keys of
    diff entries on each side will always match a boundary
    exactly. The only situation where multiple diff entries
    on one side matches a boundary is when add/remove or
    add/patch pairs occur, i.e. when inserting something
    just before an item that is removed or modified.
    """
    i_diffs = [0] * len(diffs)
    chunks = []
    nb = len(boundaries)
    for i in range(nb):
        # Find span of next chunk
        j = boundaries[i]
        k = boundaries[i+1] if i < nb-1 else j
        # Collect diff entries from each side
        # starting at beginning of this chunk
        sub_diffs = []
        for m, d in enumerate(diffs):
            dis = []
            while i_diffs[m] < len(d) and d[i_diffs[m]].key == j:
                dis += [d[i_diffs[m]]]
                i_diffs[m] += 1
            sub_diffs.append(dis)
        # Add non-empty chunks
        if j < k or any(sub_diffs):
            chunks.append((j, k) + tuple(sub_diffs))
    return chunks


def make_merge_chunks(base, *diffs, **kwargs):
    """Return list of chunks (i, j, d0, d1, ..., dn) where dX are
    lists of diff entries affecting the range base[i:j].

    If d0 and d1 are both empty the chunk is not modified.

    Includes full range 0:len(base).

    Each diff list contains either 0, 1, or 2 entries,
    in case of 2 entries the first will be an insert
    at i (the beginning of the range) and the other a
    removerange or patch covering the full range i:j.
    """
    if kwargs.get("single_item"):
        # Split diff on single items such that no chunk or
        # diff entry covers more than one base item
        boundaries = list(range(len(base)+1))
    else:
        # Split diffs on union of diff entry boundaries such that
        # no diff entry overlaps with more than one other entry.
        # Including 0,N makes loop over chunks cleaner.
        boundaries = set((0, len(base)))
        for d in diffs:
            boundaries |= get_section_boundaries(d)
        boundaries = sorted(boundaries)

    split_diffs = [split_diffs_on_boundaries(d, boundaries) for d in diffs]

    # Make list of chunks on the form (j, k, diffs)
    chunks = make_chunks(boundaries, split_diffs)

    # Some sanity checking
    if base or split_diffs:
        assert chunks, 'no merge chunks produced'
        assert chunks[0][0] == 0, 'invalid range start of first merge chunk'
        assert chunks[-1][1] == len(base), 'invalid range end of final merge chunk'

    return chunks


def chunk_typename(diffs):
    """For the diffs of a chunk, return string representations of its type.

    Returns a two-tuple of strings.
     - The first string contains zero or more characters 'a' and/or 'A', where
       'a' signifies a dict 'add' operation, and 'A' signifies a sequence
       'addrange' operation.
    - The second string contains zero or more characters 'P', 'R', 'r',
      and/or 'c': 'P': patch, 'R': removerange, 'r': remove, 'c': replace

    For a proper chunk, each string should have either 0 or 1 character each,
    but this function will not perform any checks.
    """
    aname = ""
    pname = ""
    for e in diffs:
        if e.op == DiffOp.ADDRANGE:
            aname += "A"
        elif e.op == DiffOp.ADD:
            aname += "a"
        elif e.op == DiffOp.PATCH:
            pname += "P"
        elif e.op == DiffOp.REMOVERANGE:
            pname += "R"
        elif e.op == DiffOp.REMOVE:
            pname += "r"
        elif e.op == DiffOp.REPLACE:
            pname += "c"
    return aname, pname




import os


def test_nbdime_js_artifacts_present():
    from nbdime.webapp import __file__ as webapp_init
    static_folder = os.path.join(os.path.dirname(webapp_init), 'static')
    targets = [os.path.join(static_folder, 'nbdime.js')]
    missing = [t for t in targets if not os.path.exists(t)]
    if missing:
        raise ValueError(('missing files: %s' % missing))



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from .generic import decide_merge
from .decisions import apply_decisions
from .notebooks import merge_notebooks

__all__ = ["decide_merge", "merge_notebooks", "apply_decisions"]



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.




from contextlib import contextmanager
import os
import requests
import shlex
import sys
import time
from subprocess import check_output, check_call, STDOUT, CalledProcessError

import pytest
import nbformat

from nbdime import patch, diff
from nbdime.diff_format import is_valid_diff


from shutil import which

pjoin = os.path.join

have_git = which('git')
have_hg = which('hg')

WEB_TEST_TIMEOUT = 15

TEST_TOKEN = 'nbdime-test-token'


def assert_is_valid_notebook(nb):
    """These are the current assumptions on notebooks in these tests. Loosen on demand."""
    assert nb["nbformat"] == 4
    # assert nb["nbformat_minor"] == 0
    assert isinstance(nb["metadata"], dict)
    assert isinstance(nb["cells"], list)
    assert all(isinstance(cell, dict) for cell in nb["cells"])


def check_diff_and_patch(a, b):
    "Check that patch(a, diff(a,b)) reproduces b."
    d = diff(a, b)
    assert is_valid_diff(d)
    assert patch(a, d) == b


def check_symmetric_diff_and_patch(a, b):
    "Check that patch(a, diff(a,b)) reproduces b and vice versa."
    check_diff_and_patch(a, b)
    check_diff_and_patch(b, a)


def sources_to_notebook(sources, cell_type='code', id_seed=0, strip_ids=False):
    assert isinstance(sources, list)
    nb = nbformat.v4.new_notebook()
    for source in sources:
        if isinstance(source, list):
            source = "".join(source)
        if cell_type == 'code':
            nb.cells.append(nbformat.v4.new_code_cell(source))
        elif cell_type == 'markdown':
            nb.cells.append(nbformat.v4.new_markdown_cell(source))
    if strip_ids:
        strip_cell_ids(nb)
    else:
        deterministic_cell_ids(nb)
    return nb


def notebook_to_sources(nb, as_str=True):
    sources = []
    for cell in nb["cells"]:
        source = cell["source"]
        if as_str and isinstance(source, list):
            source = "\n".join([line.strip("\n") for line in source])
        elif not as_str and isinstance(source, str):
            source = source.splitlines(True)
        sources.append(source)
    return sources


def outputs_to_notebook(outputs, strip_ids=False):
    assert isinstance(outputs, list)
    assert len(outputs) == 0 or isinstance(outputs[0], list)
    nb = nbformat.v4.new_notebook()
    for cell_outputs in outputs:
        cell = nbformat.v4.new_code_cell()
        nb.cells.append(cell)
        for output in cell_outputs:
            if isinstance(output, str):
                output = nbformat.v4.new_output(
                    output_type="display_data",
                    data={
                        "text/plain": output,
                    }
                )
            assert isinstance(output, dict)
            cell.outputs.append(output)
    if strip_ids:
        strip_cell_ids(nb)
    else:
        deterministic_cell_ids(nb)
    return nb


def strip_cell_id(cell):
    if "id" in cell:
        del cell["id"]
    return cell


def strip_cell_ids(nb):
    for cell in nb["cells"]:
        strip_cell_id(cell)
    return nb


def deterministic_cell_ids(nb):
    for i, cell in enumerate(nb["cells"]):
        cell["id"] = f"cell-id-{i}"
    return nb

def new_cell_wo_id(*args, **kwargs):
    cell = nbformat.v4.new_code_cell(*args, **kwargs)
    strip_cell_id(cell)
    return cell


@contextmanager
def assert_clean_exit():
    """Assert that SystemExit is called with code=0"""
    with pytest.raises(SystemExit) as e:
        yield
    assert e.value.code == 0


def call(cmd):
    """Call a command

    if str, split into command list
    """
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    return check_call(cmd, stdout=sys.stdout, stderr=sys.stderr)


def get_output(cmd, err=False, returncode=0):
    """Run a command and get its output (as text)"""
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    stderr = STDOUT if err else sys.stderr
    try:
        output = check_output(cmd, stderr=stderr).decode('utf8', 'replace')
    except CalledProcessError as e:
        if e.returncode != returncode:
            raise
        return e.output.decode('utf8', 'replace')
    else:
        if returncode != 0:
            raise CalledProcessError(0, cmd, output.encode('utf8'), stderr)
    return output


def wait_up(url, interval=0.1, check=None):
    while True:
        try:
            r = requests.get(url)
        except Exception:
            if check:
                assert check()
            print("waiting for %s" % url)
            time.sleep(interval)
        else:
            break



MERCURIAL_TEST_CONFIG = """
[extensions]
extdiff =

[extdiff]
cmd.nbdiff = hg-nbdiff
cmd.nbdiffweb = hg-nbdiffweb
opts.nbdiffweb = --log-level DEBUG --browser=disabled

[merge-tools]
nbdime.priority = 2
nbdime.premerge = False
nbdime.executable = hg-nbmerge
nbdime.args = $base $local $other $output
nbdimeweb.priority = 1
nbdimeweb.premerge = False
nbdimeweb.executable = hg-nbmergeweb
nbdimeweb.args = --log-level DEBUG --browser=disabled $base $local $other $output
nbdimeweb.gui = True

[merge-patterns]
**.ipynb = nbdime
"""


def write_local_hg_config(repo_path):
    fn = pjoin(repo_path, ".hg", "hgrc")
    with open(fn, 'w') as f:
        f.write(MERCURIAL_TEST_CONFIG)



# -*- coding: utf-8 -*-

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import shutil

from nbdime.diffing.directorydiff import diff_directories
from nbdime.utils import EXPLICIT_MISSING_FILE

pjoin = os.path.join
relpath = os.path.relpath



def make_dirs(tmpdir):
    tmpdir.mkdir('dir1')
    tmpdir.mkdir('dir2')

    # Set up identical directory structures in both
    tmpdir.mkdir('dir1', 'subA')
    tmpdir.mkdir('dir2', 'subA')

    tmpdir.mkdir('dir1', 'subA', 'subsub')
    tmpdir.mkdir('dir2', 'subA', 'subsub')

    tmpdir.mkdir('dir1', 'subB')
    tmpdir.mkdir('dir2', 'subB')


def test_diff_root_files(tmpdir, filespath):
    src = filespath

    make_dirs(tmpdir)

    dir1 = str(tmpdir.join('dir1'))
    dir2 = str(tmpdir.join('dir2'))

    shutil.copy(pjoin(src, 'src-and-output--1.ipynb'), pjoin(dir1, 'a.ipynb'))
    shutil.copy(pjoin(src, 'src-and-output--2.ipynb'), pjoin(dir2, 'a.ipynb'))

    diffs = diff_directories(dir1, dir2)
    # Make paths relative for comparison:
    diffs = tuple((relpath(a, dir1), relpath(b, dir2)) for (a, b) in diffs)
    assert diffs == (('a.ipynb', 'a.ipynb'),)


def test_diff_sub_files(tmpdir, filespath):
    src = filespath

    make_dirs(tmpdir)

    dir1 = str(tmpdir.join('dir1'))
    dir2 = str(tmpdir.join('dir2'))

    shutil.copy(pjoin(src, 'src-and-output--1.ipynb'), pjoin(dir1, 'subA', 'subsub', 'a.ipynb'))
    shutil.copy(pjoin(src, 'src-and-output--2.ipynb'), pjoin(dir2, 'subA', 'subsub', 'a.ipynb'))

    diffs = diff_directories(dir1, dir2)
    # Make paths relative for comparison:
    diffs = tuple((relpath(a, dir1), relpath(b, dir2)) for (a, b) in diffs)
    assert diffs == ((pjoin('subA', 'subsub', 'a.ipynb'), pjoin('subA', 'subsub', 'a.ipynb')),)


def test_diff_file_added(tmpdir, filespath):
    src = filespath

    make_dirs(tmpdir)

    dir1 = str(tmpdir.join('dir1'))
    dir2 = str(tmpdir.join('dir2'))

    shutil.copy(pjoin(src, 'src-and-output--2.ipynb'), pjoin(dir2, 'subA', 'subsub', 'a.ipynb'))

    diffs = diff_directories(dir1, dir2)
    # Make paths relative for comparison:
    diffs = tuple((a, relpath(b, dir2)) for (a, b) in diffs)
    assert diffs == ((EXPLICIT_MISSING_FILE, pjoin('subA', 'subsub', 'a.ipynb')),)


def test_diff_file_deleted(tmpdir, filespath):
    src = filespath

    make_dirs(tmpdir)

    dir1 = str(tmpdir.join('dir1'))
    dir2 = str(tmpdir.join('dir2'))

    shutil.copy(pjoin(src, 'src-and-output--1.ipynb'), pjoin(dir1, 'subA', 'subsub', 'a.ipynb'))

    diffs = diff_directories(dir1, dir2)
    # Make paths relative for comparison:
    diffs = tuple((relpath(a, dir1), b) for (a, b) in diffs)
    assert diffs == ((pjoin('subA', 'subsub', 'a.ipynb'), EXPLICIT_MISSING_FILE),)




# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import argparse
import json
import logging
import os
import sys

from ._version import __version__
from .config import (
    get_defaults_for_argparse, build_config, entrypoint_configurables,
    Namespace
)
from .diffing.notebooks import set_notebook_diff_targets, set_notebook_diff_ignores
from .gitfiles import is_gitref
from .ignorables import diff_ignorables
from .log import init_logging, set_nbdime_log_level


class ConfigBackedParser(argparse.ArgumentParser):

    def parse_known_args(self, args=None, namespace=None):
        entrypoint = self.prog.split(' ')[0]
        try:
            defs = get_defaults_for_argparse(entrypoint)
            ignore = defs.pop('Ignore', None)
            self.set_defaults(**defs)
            if ignore:
                set_notebook_diff_ignores(ignore)
        except ValueError:
            pass
        return super(ConfigBackedParser, self).parse_known_args(args=args, namespace=namespace)


class LogLevelAction(argparse.Action):
    def __init__(self, option_strings, dest, default=None, **kwargs):
        # __call__ is not called if option not given:
        level = getattr(logging, default or 'INFO')
        init_logging(level=level)
        set_nbdime_log_level(level)
        super(LogLevelAction, self).__init__(option_strings, dest, default=default, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
        level = getattr(logging, values)
        set_nbdime_log_level(level, True)


class SkipAction(argparse.Action):
    """Action of an argument that will not be stored"""
    def __init__(self, option_strings, dest, **kwargs):
        super(SkipAction, self).__init__([], argparse.SUPPRESS, **kwargs)

    def __call__(self, parser, ns, values, opttion_string=None):
        pass


class PathType(object):
    """Argparse type for arguments that should be paths

    No-op on Python 3, but casts Python 2 bytes to text
    using sys.getfilesystemencoding()
    """
    def __init__(self):
        pass

    def __call__(self, value):
        if not isinstance(value, bytes):
            return value

        # py2: decode bytes to text
        encoding = sys.getfilesystemencoding() or 'utf-8'
        if encoding.lower() == 'ascii':
            # ignore ascii and use utf-8
            # if it really is ascii, this will still be correct,
            # but it never should actually be ascii
            encoding = 'utf-8'

        return value.decode(encoding)


Path = PathType()


def modify_config_for_print(config):
    output = {}
    ns = None
    for k, v in config.items():
        if isinstance(v, dict):
            output[k] = modify_config_for_print(v)
            if not output[k]:
                output[k] = '{}'
        elif k in diff_ignorables and v is None:
            if ns is None:
                ns = Namespace(config)
                for k2 in diff_ignorables:
                    setattr(ns, k2, config.get(k2, None))
                process_exclusive_ignorables(ns, diff_ignorables)
            output[k] = '<unset, resolves to {0}>'.format(
                json.dumps(getattr(ns, k, v)))
        else:
            output[k] = json.dumps(v)
    return output


class ConfigHelpAction(argparse.Action):
    def __init__(self, option_strings, dest, help=None):
        super(ConfigHelpAction, self).__init__(
            option_strings, dest, nargs=0, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        from .prettyprint import pretty_print_dict, PrettyPrintConfig

        header = entrypoint_configurables[parser.prog].__name__
        config = build_config(parser.prog, True)
        pretty_print_dict(
            {
                header: modify_config_for_print(config),
            },
            config=PrettyPrintConfig(out=sys.stderr)
        )
        sys.exit(1)


class IgnorableAction(argparse.Action):
    """Adds the supplied positive options and negative/ignore version as well"""

    def __init__(self, option_strings, dest, default=None, required=False, help=None):
        opts = []
        for opt in option_strings:
            if len(opt) == 2 and opt[0] == '-':
                if not opt[1].islower():
                    raise ValueError('Single character flags should be lower-case for IgnorableAction')
                opts.append(opt)
                opts.append(opt.upper())
            elif opt[:2] == '--':
                opts.append(opt)
                opts.append('--ignore-' + opt[2:])
            else:
                ValueError('Could not turn option "%s" into an IgnoreAction option.')

        # Put positives first, negatives last:
        opts = opts[0::2] + opts[1::2]

        super(IgnorableAction, self).__init__(
            opts, dest, nargs=0, const=None,
            default=default, required=required,
            help=help)

    def __call__(self, parser, ns, values, option_string=None):
        if len(option_string) == 2:
            setattr(ns, self.dest, option_string[1].islower())
        else:
            setattr(ns, self.dest, option_string[2 : 2 + len('ignore')] != 'ignore')


def process_exclusive_ignorables(ns, arg_names, default=True):
    """Parse a set of ignorables.

    It checks that all specified options are either all positive or all negative.
    It then returns a namespace with the parsed options.

    Returns whether any values were specified or not.
    """
    # `toggle` tracks whether:
    #  - True: One or more positive options were defined
    #  - False: One or more negative options were defined
    #  - None: No options were defined
    toggle = getattr(ns, arg_names[0])
    for name in arg_names[1:]:
        opt = getattr(ns, name)
        if toggle is None:
            toggle = opt
        elif toggle != opt and opt is not None:
            message = 'Arguments must either all be negative or all positive: %r' % (arg_names,)
            raise argparse.ArgumentError(None, message)

    if toggle is not None:
        # One or more options were defined, set default to the opposite
        default = not toggle

    # Set all unset options to the default
    for name in arg_names:
        if getattr(ns, name) is None:
            setattr(ns, name, default)
    return toggle is not None


def add_generic_args(parser):
    """Adds a set of arguments common to all nbdime commands.
    """
    parser.add_argument(
        '--version',
        action="version",
        version="%(prog)s " + __version__)
    parser.add_argument(
        '--config',
        help="list the valid config keys and their current effective values",
        action=ConfigHelpAction,
    )
    parser.add_argument(
        '--log-level',
        default='INFO',
        choices=('DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'),
        help="set the log level by name.",
        action=LogLevelAction,
    )


def add_git_config_subcommand(subparsers, enable, disable, subparser_help, enable_help, disable_help):
    # Add subparser
    config = subparsers.add_parser('config',
        description=subparser_help)

    # Option for git scope (global/system):
    scope = config.add_mutually_exclusive_group()
    scope.add_argument('--global', action='store_const', dest='scope',
        const='global',
        help="configure your global git config instead of the current repo"
    )
    scope.add_argument('--system', action='store_const', dest='scope',
        const='system',
        help="configure your system git config instead of the current repo"
    )

    # Add enable/disable flags
    enable_disable = config.add_mutually_exclusive_group(required=True)
    enable_disable.add_argument('--enable', action='store_const',
        dest='config_func', const=enable,
        help=enable_help
    )
    enable_disable.add_argument('--disable', action='store_const',
        dest='config_func', const=disable,
        help=disable_help
    )
    return config


def add_web_args(parser, default_port=8888):
    """Adds a set of arguments common to all commands that show a web gui.
    """
    port_help = (
        "specify the port you want the server to run on. Default is %d%s." % (
            default_port, " (random)" if default_port == 0 else ""
        ))
    parser.add_argument(
        '-p', '--port',
        default=default_port,
        type=int,
        help=port_help)
    parser.add_argument(
        '-b', '--browser',
        default=None,
        type=str,
        help="specify the browser to use, to override the system default.")
    parser.add_argument(
        '--persist',
        action="store_true",
        default=False,
        help="prevent server shutting down on remote close request (when these"
             " would normally be supported)."
    )
    parser.add_argument(
        '--ip',
        default='127.0.0.1',
        help="specify the interface to listen to for the web server. "
        "NOTE: Setting this to anything other than 127.0.0.1/localhost "
        "might comprimise the security of your computer. Use with care!")
    cwd = os.path.abspath(os.path.curdir)

    parser.add_argument(
        '-w', '--workdirectory',
        default=cwd,
        help="specify the working directory you want "
             "the server to run from. Default is the "
             "actual cwd at program start.")
    parser.add_argument(
        '--base-url',
        default='/',
        help="The base URL prefix under which to run the web app")
    parser.add_argument(
        '--show-unchanged',
        dest='hide_unchanged',
        action="store_false",
        default=True,
        help="show unchanged cells by default"
    )
    parser.add_argument(
        '--identical-lines-margin',
        dest='identical_lines_margin',
        default=2,
        type=int,
        help="Margin for collapsing identical lines in editor; set to -1 to deactivate.",
    )


def add_diff_args(parser):
    """Adds a set of arguments for commands that perform diffs.

    Note:
        Merge applications also performs diff operations to compute
        the merge, so these arguments should also be included there.
    """
    # TODO: Define sensible strategy variables and implement
    #parser.add_argument('-X', '--diff-strategy',
    #                    default="default", choices=("foo", "bar"),
    #                    help="specify the diff strategy to use.")

    # Things we can choose to diff or not
    ignorables = parser.add_argument_group(
        title='ignorables',
        description='Set which parts of the notebook (not) to process.')
    ignorables.add_argument(
        '-s', '--sources',
        action=IgnorableAction,
        help="process/ignore sources.")
    ignorables.add_argument(
        '-o', '--outputs',
        action=IgnorableAction,
        help="process/ignore outputs.")
    ignorables.add_argument(
        '-a', '--attachments',
        action=IgnorableAction,
        help="process/ignore attachments.")
    ignorables.add_argument(
        '-m', '--metadata',
        action=IgnorableAction,
        help="process/ignore metadata.")
    ignorables.add_argument(
        '-i', '--id',
        action=IgnorableAction,
        help="process/ignore identifiers.")
    ignorables.add_argument(
        '-d', '--details',
        action=IgnorableAction,
        help="process/ignore details not covered by other options.")


def add_diff_cli_args(parser):
    """Adds a set of arguments for CLI diff commands (i.e. not web).
    """
    parser.add_argument(
        '--color-words',
        action='store_true', default=False,
        help=("whether to pass the --color-words flag to any internal calls "
              "to git diff")
    )


def add_filter_args(diff_parser):
    """Adds configuration for git commands where filter use is flagged"""
    # Ideally, we would want to apply the filter only if we knew
    # the file was not from a blob. However, this is not possible:
    # If remote file path is equal to repo file path, it implies
    # that the hex of remote equals the hex of the file on disk.
    # Two possible cases can cause this:
    # 1) Diffing against working dir (or stage when entire file is staged)
    # 2) Diffing against a blob (clean) that happens to have the same hex as
    #    the (smudged) file in working tree.
    # Condition 1 should have filter applied, 2 should not.
    # We can learn something by comparing the remote hash to the hash of the
    # file in HEAD.
    # - If they are equal, we know that is cannot come from a diff
    #   agains working tree (git would not see it as changed),
    #   so it must be from a blob (clean). No filter.
    # - If they differ, consider the setup:
    #   git co A; git co B -- file.path; git reset A
    #   + remote could be from a working-tree diff: git diff (smudged, apply filter).
    #   + remote could be from a blob: git diff A B (clean, no filter).
    #
    # These are undistinguishable to us. Therefore, we will always
    # apply the filter to the remote file if flag use_filter is set.
    diff_parser.add_argument(
        '--use-filter',
        action='store_true', default=False,
        help='apply any configured git filters on remote')


def add_git_diff_driver_args(diff_parser):
    """Adds a set of 7 stanard git diff driver arguments:
        path old-file old-hex old-mode new-file new-hex new-mode [ rename-to rename-metadata ]

    Note: Only path, base and remote are added to parsed namespace
    """
    add_filter_args(diff_parser)
    diff_parser.add_argument('path', type=Path)
    diff_parser.add_argument('base', type=Path, nargs='?', default=None)
    diff_parser.add_argument('base_sha1', nargs='?', default=None, action=SkipAction)
    diff_parser.add_argument('base_mode', nargs='?', default=None, action=SkipAction)
    diff_parser.add_argument('remote', type=Path, nargs='?', default=None)
    diff_parser.add_argument('remote_sha1', nargs='?', default=None, action=SkipAction)
    diff_parser.add_argument('remote_mode', nargs='?', default=None, action=SkipAction)
    diff_parser.add_argument('rename_to', type=Path, nargs='?', default=None, action=SkipAction)
    diff_parser.add_argument('rename_metadata', type=Path, nargs='?', default=None, action=SkipAction)


def process_diff_flags(args):
    any_flags_given = process_exclusive_ignorables(args, diff_ignorables)
    if any_flags_given:
        # Note: This will blow away any options set via config (for these fields)
        set_notebook_diff_targets(
            args.sources, args.outputs, args.attachments, args.metadata,
            args.id, args.details)


def resolve_diff_args(args):
    """Resolve ambiguity of path vs base/remote for git:

    Cases:
     - No args: Use defaults
     - One arg: Either base or path, check with is_gitref.
     - Two args or more: Check if first two are base/remote by is_gitref
    """
    base = args.base
    remote = args.remote
    paths = getattr(args, 'paths', None)
    if not paths:
        paths = None
    if remote is None and paths is None:
        # One arg only:
        if not is_gitref(base):
            paths = base
            base = 'HEAD'
    # Two or more args:
    elif paths is None:
        # Two exactly
        # - Two files (not git-mode, do nothing)
        # - Base gitref one file (remote=None, path = file)
        # - Base gitref remote gitref (do nothing)
        if is_gitref(base) and not is_gitref(remote):
            paths = remote
            remote = None
    elif base and remote:
        # Three or more
        if not is_gitref(base):
            paths = [base, remote] + paths
            base = remote = None
        elif is_gitref(base) and not is_gitref(remote):
            paths = [remote] + paths
            remote = None
    return base, remote, paths


def add_merge_args(parser):
    """Adds a set of arguments for commands that perform merges.
    """
    from .merging.notebooks import cli_conflict_strategies, cli_conflict_strategies_input, cli_conflict_strategies_output
    parser.add_argument(
        '--merge-strategy',
        default="inline",
        choices=cli_conflict_strategies,
        help="the merge strategy to use.")
    parser.add_argument(
        '--input-strategy',
        default=None,
        choices=cli_conflict_strategies_input,
        help="the merge strategy to use for inputs "
             "(overrides 'merge-strategy' for inputs).")
    parser.add_argument(
        '--output-strategy',
        default=None,
        choices=cli_conflict_strategies_output,
        help="the merge strategy to use for outputs "
             "(overrides 'merge-strategy' for outputs).")
    parser.add_argument(
        '--no-ignore-transients',
        dest='ignore_transients',
        action="store_false",
        default=True,
        help="disallow deletion of transient data such as outputs and "
             "execution counts in order to resolve conflicts.")


filename_help = {
    "notebook": "The notebook filename.",
    "base":   "The base notebook filename.",
    "local":  "The local modified notebook filename.",
    "remote": "The remote modified notebook filename.",
    "merged": "The merge result notebook filename.",
    "patch":  "The patch filename, output from nbdiff.",
    }


def add_filename_args(parser, names):
    """Add the base, local, remote, and merged positional arguments.

    Helps getting consistent doc strings.
    """
    for name in names:
        parser.add_argument(name, type=Path, help=filename_help[name])


def add_prettyprint_args(parser):
    """Adds optional arguments for controlling pretty print behavior.
    """
    parser.add_argument(
        '--no-color',
        dest='use_color',
        action="store_false",
        default=True,
        help=("prevent use of ANSI color code escapes for text output")
    )
    parser.add_argument(
        '--no-git',
        dest='use_git',
        action="store_false",
        default=True,
        help=("prevent use of git for formatting diff/merge text output")
    )
    parser.add_argument(
        '--no-use-diff',
        dest='use_diff',
        action="store_false",
        default=True,
        help=("prevent use of diff/diff3 for formatting diff/merge text output")
    )


def prettyprint_config_from_args(arguments, **kwargs):
    from .prettyprint import PrettyPrintConfig
    return PrettyPrintConfig(
        include=arguments,
        color_words=getattr(arguments, 'color_words', False),
        use_color=getattr(arguments, 'use_color', True),
        use_git=getattr(arguments, 'use_git', True),
        use_diff=getattr(arguments, 'use_diff', True),
        **kwargs
    )


def args_for_server(arguments):
    """Translate standard arguments into kwargs for running webapp.nbdimeserver.main"""
    # Map format: <arguments.name>='<kwargs[key]>'
    kmap = dict(ip='ip',
                port='port',
                workdirectory='cwd',
                base_url='base_url',
                hide_unchanged='hide_unchanged',
                identical_lines_margin='identical_lines_margin',
                )
    ret = {kmap[k]: v for k, v in vars(arguments).items() if k in kmap}
    if 'persist' in arguments:
        ret['closable'] = not arguments.persist
    return ret


def args_for_browse(arguments):
    """Translate standard arguments into kwargs for webapp.webutil.browse()"""
    # Map format: <arguments.name>='<kwargs[key]>'
    kmap = dict(ip='ip',
                browser='browsername',
                base_url='base_url',
                )
    return {kmap[k]: v for k, v in vars(arguments).items() if k in kmap}



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import base64
import io
import json
import logging
import os
import sys

from jinja2 import FileSystemLoader, Environment
import nbformat
from jupyter_server.base.handlers import JupyterHandler, APIHandler
from jupyter_server.utils import url_path_join
from jupyter_server.log import log_request
import requests
from tornado import ioloop, web, escape, netutil, httpserver

from .. import __file__ as nbdime_root
from ..args import ConfigBackedParser, add_generic_args, add_web_args
from ..diffing.notebooks import diff_notebooks
from ..log import logger
from ..merging.notebooks import decide_notebook_merge
from ..nbmergeapp import _build_arg_parser as build_merge_parser
from ..utils import EXPLICIT_MISSING_FILE, is_in_repo


# TODO: See <notebook>/notebook/services/contents/handlers.py for possibly useful utilities:
#contents_manager
#ContentsHandler


# Separate logger for server entrypoint (?)
_logger = logging.getLogger(__name__)


here = os.path.abspath(os.path.dirname(__file__))
static_path = os.path.join(here, 'static')
template_path = os.path.join(here, 'templates')


class NbdimeHandler(JupyterHandler):
    def initialize(self, **params):
        self.params = params

    def base_args(self):
        fn = self.params.get('outputfilename', None)
        base = {
            'closable': self.params.get('closable', False),
            'savable': fn is not None,
            'baseUrl': self.nbdime_base_url,
            'hideUnchanged': self.params.get('hide_unchanged', True),
            'collapseIdentical': self.params.get('identical_lines_margin', 2),
        }
        if fn:
            # For reference, e.g. if user wants to download file
            base['outputfilename'] = fn
        return base

    def write_error(self, status_code, **kwargs):
        # Write error message for HTTPErrors if serve_traceback is off:
        exc_info = kwargs.get('exc_info', None)
        if exc_info and not self.settings.get('serve_traceback'):
            (etype, value, traceback) = exc_info
            if etype == web.HTTPError:
                self.set_header('Content-Type', 'text/plain')
                return self.finish(str(value))
        return super(NbdimeHandler, self).write_error(status_code, **kwargs)

    def read_notebook(self, arg, fail_on_empty=True):
        # Currently assuming arg is a filename relative to
        # where the server was started from, later we may
        # want to accept urls or full notebooks as well.
        if not isinstance(arg, str):
            raise web.HTTPError(400, 'Expecting a filename or a URL.')

        try:
            # Check that file exists
            if arg == EXPLICIT_MISSING_FILE:
                path = arg
            else:
                path = os.path.join(self.curdir, arg)
                if not os.path.exists(path):
                    if '://' not in arg:
                        raise ValueError('Supplied argument cannot be read: %r' % arg)
                    # Assume file is URI
                    r = requests.get(arg)
                    r.raise_for_status()

            # Let nbformat do the reading and validation
            if path == EXPLICIT_MISSING_FILE:
                nb = nbformat.v4.new_notebook()
            elif os.path.exists(path):
                try:
                    nb = nbformat.read(path, as_version=4)
                except nbformat.reader.NotJSONError:
                    if fail_on_empty:
                        raise
                    # Handle empty notebook file
                    if isinstance(path, str):
                        with io.open(path, encoding='utf-8') as fo:
                            if len(fo.read(10)) != 0:
                                raise
                    nb = nbformat.v4.new_notebook()
            else:
                nb = nbformat.reads(r.text, as_version=4)
        except requests.exceptions.HTTPError as e:
            self.log.exception(e)
            raise web.HTTPError(422, 'Invalid notebook: %s, received http error: %s' % (arg, str(e)))
        except Exception as e:
            self.log.exception(e)
            raise web.HTTPError(422, 'Invalid notebook: %s' % arg)

        return nb

    def get_notebook_argument(self, argname):
        # Assuming a request on the form "{'argname':arg}"
        body = json.loads(escape.to_unicode(self.request.body))
        arg = body[argname]

        return self.read_notebook(arg)

    @property
    def nbdime_base_url(self):
        relative = self.params.get('nbdime_relative_base_url', None)
        if relative is None:
            return self.base_url
        return url_path_join(self.base_url, relative)

    @property
    def curdir(self):
        return self.params.get('cwd', os.curdir)

    # Polyfill in until we can rely on Notebook 5.0 being present
    @property
    def mathjax_config(self):
        try:
            return super(NbdimeHandler, self).mathjax_config
        except AttributeError:
            return 'TeX-AMS_HTML-full,Safe'


class MainHandler(NbdimeHandler):
    def get(self):
        args = self.base_args()
        args['base'] = self.get_argument('base', '')
        args['local'] = self.get_argument('local', '')
        args['remote'] = self.get_argument('remote', '')
        self.write(self.render_template('compare.html',
                    config_data=args,
                   ))


class MainDiffHandler(NbdimeHandler):
    def get(self):
        args = self.base_args()
        args['base'] = self.get_argument('base', '')
        args['remote'] = self.get_argument('remote', '')

        self.write(self.render_template('diff.html',
                    config_data=args,
                   ))


class MainDifftoolHandler(NbdimeHandler):
    def get(self):
        args = self.base_args()
        if 'difftool_args' in self.params:
            base = self.params['difftool_args']['base']
            remote = self.params['difftool_args']['remote']
            if isinstance(base, str):
                args['base'] = base
            else:
                args['base'] = base.name
            if isinstance(remote, str):
                args['remote'] = remote
            else:
                args['remote'] = remote.name
        else:
            args['base'] = self.get_argument('base', '')
            args['remote'] = self.get_argument('remote', '')
        self.write(self.render_template('difftool.html',
                    config_data=args,
                   ))


class MainMergeHandler(NbdimeHandler):
    def get(self):
        args = self.base_args()
        args['base'] = self.get_argument('base', '')
        args['local'] = self.get_argument('local', '')
        args['remote'] = self.get_argument('remote', '')
        args['showBase'] = self.params.get('show_base', True)
        self.write(self.render_template('merge.html',
                    config_data=args,
                   ))


class MainMergetoolHandler(NbdimeHandler):
    def get(self):
        args = self.base_args()
        if 'mergetool_args' in self.params:
            args['base'] = self.params['mergetool_args']['base']
            args['local'] = self.params['mergetool_args']['local']
            args['remote'] = self.params['mergetool_args']['remote']
            args['showBase'] = self.params.get('show_base', True)
        else:
            args['base'] = self.get_argument('base', '')
            args['local'] = self.get_argument('local', '')
            args['remote'] = self.get_argument('remote', '')
            args['showBase'] = self.params.get('show_base', True)
        self.write(self.render_template('mergetool.html',
                    config_data=args,
                   ))


class ApiDiffHandler(NbdimeHandler, APIHandler):
    def post(self):
        base_nb = self.get_notebook_argument('base')
        remote_nb = self.get_notebook_argument('remote')

        try:
            thediff = diff_notebooks(base_nb, remote_nb)
        except Exception:
            logger.exception('Error diffing documents:')
            raise web.HTTPError(500, 'Error while attempting to diff documents')

        data = {
            'base': base_nb,
            'diff': thediff,
            }
        self.finish(data)

    def get_notebook_argument(self, argname):
        if 'difftool_args' in self.params:
            arg = self.params['difftool_args'][argname]
            if not isinstance(arg, str):
                # Assume arg is file-like
                arg.seek(0)
                return nbformat.read(arg, as_version=4)
            return self.read_notebook(arg)
        return super(ApiDiffHandler, self).get_notebook_argument(argname)


class ApiMergeHandler(NbdimeHandler, APIHandler):
    def post(self):
        base_nb = self.get_notebook_argument('base')
        local_nb = self.get_notebook_argument('local')
        remote_nb = self.get_notebook_argument('remote')
        merge_args = self.settings.get('merge_args')
        if merge_args is None:
            merge_args = build_merge_parser().parse_args(['', '', ''])
            merge_args.merge_strategy = 'mergetool'
            self.settings['merge_args'] = merge_args

        try:
            decisions = decide_notebook_merge(base_nb, local_nb, remote_nb,
                                              args=merge_args)
        except Exception:
            logger.exception('Error merging documents:')
            raise web.HTTPError(500, 'Error while attempting to merge documents')

        data = {
            'base': base_nb,
            'merge_decisions': decisions
            }
        self.finish(data)

    def get_notebook_argument(self, argname):
        if 'mergetool_args' in self.params:
            arg = self.params['mergetool_args'][argname]
            return self.read_notebook(arg, fail_on_empty=False)
        return super(ApiMergeHandler, self).get_notebook_argument(argname)


class ApiMergeStoreHandler(NbdimeHandler, APIHandler):
    def post(self):
        # I don't think we want to accept arbitrary filenames
        # to write to from the http request, only allowing
        # this operation if the server was run with an output
        # filename as a commandline argument:
        fn = self.params.get('outputfilename', None)
        if not fn:
            raise web.HTTPError(400, 'Server does not accept storing merge result.')
        path = os.path.join(self.curdir, fn)
        logger.info('Saving merge result in %s', path)

        body = json.loads(escape.to_unicode(self.request.body))
        merged = body['merged']
        merged_nb = nbformat.from_dict(merged)

        # Somehow store unsolved conflicts?
        # conflicts = body['conflicts']

        with io.open(path, 'w', encoding='utf8') as f:
            nbformat.write(merged_nb, f)
        self.finish()


class ApiCloseHandler(NbdimeHandler, APIHandler):
    def post(self):
        # Only allow closing, if started as tool
        if self.params.get('closable', False) is not True:
            raise web.HTTPError(
                400, 'This server cannot be closed remotely.')

        # Fail if no exit code is supplied:
        fallback = int(self.request.headers.get('exit_code', 1))
        try:
            self.application.exit_code = self.get_argument('exitCode')
        except web.MissingArgumentError:
            try:
                self.application.exit_code = json.loads(self.request.body).get('exitCode', fallback)
            except json.JSONDecodeError:
                self.application.exit_code = fallback

        if isinstance(self.application.exit_code, str):
            self.application.exit_code = int(self.application.exit_code, 10)

        _logger.info('Closing server on remote request (%d)', self.application.exit_code)
        self.finish()
        ioloop.IOLoop.current().stop()


def asyncio_patch():
    """set default asyncio policy to be compatible with tornado

    Tornado 6 (at least) is not compatible with the default
    asyncio implementation on Windows

    Pick the older SelectorEventLoopPolicy on Windows
    if the known-incompatible default policy is in use.

    do this as early as possible to make it a low priority and overrideable

    ref: https://github.com/tornadoweb/tornado/issues/2608

    FIXME: if/when tornado supports the defaults in asyncio,
            remove and bump tornado requirement for py38
    """
    if sys.platform.startswith("win") and sys.version_info >= (3, 8):
        import asyncio
        try:
            from asyncio import (
                WindowsProactorEventLoopPolicy,
                WindowsSelectorEventLoopPolicy,
            )
        except ImportError:
            pass
            # not affected
        else:
            if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                # fallback to the pre-3.8 default of Selector
                asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def make_app(**params):
    base_url = params.pop('base_url', '/')
    handlers = [
        (r'/', MainHandler, params),
        (r'/diff', MainDiffHandler, params),
        (r'/difftool', MainDifftoolHandler, params),
        (r'/merge', MainMergeHandler, params),
        (r'/mergetool', MainMergetoolHandler, params),
        (r'/api/diff', ApiDiffHandler, params),
        (r'/api/merge', ApiMergeHandler, params),
        (r'/api/store', ApiMergeStoreHandler, params),
        (r'/api/closetool', ApiCloseHandler, params),
        # Static handler will be added automatically
    ]
    if base_url != '/':
        prefix = base_url.rstrip('/')
        handlers = [
            (prefix + path, cls, params)
            for (path, cls, params) in handlers
        ]
    else:
        prefix = ''

    env = Environment(loader=FileSystemLoader([template_path]), autoescape=False)
    settings = {
        'log_function': log_request,
        'static_path': static_path,
        'static_url_prefix': prefix + '/static/',
        'template_path': [template_path],
        'base_url': base_url,
        'jinja2_env': env,
        'local_hostnames': ['localhost', '127.0.0.1'],
        'cookie_secret': base64.encodebytes(os.urandom(32)), # Needed even for an unsecured server.
    }

    try:
        from jupyter_server.auth import IdentityProvider
        settings['identity_provider'] = IdentityProvider()
    except ImportError:
        pass

    if is_in_repo(nbdime_root):
        # don't cache when working from repo
        settings.update({
            # 'autoreload': True,
            'compiled_template_cache': False,
            'static_hash_cache': False,
            # 'serve_traceback': True,
            })

    app = web.Application(handlers, **settings)
    app.exit_code = 0
    return app

def init_app(on_port=None, closable=False, **params):
    asyncio_patch()
    _logger.debug('Using params: %s', params)
    params.update({'closable': closable})
    port = params.pop('port', 0)
    ip = params.pop('ip', '127.0.0.1')
    app = make_app(**params)
    if ip not in {'127.0.0.1', 'localhost', '::1'}:
        # enable remote access when listening on a public ip
        app.settings['allow_remote_access'] = True
    if port != 0:
        server = app.listen(port, address=ip)
        _logger.info('Listening on %s, port %d', ip, port)
    else:
        sockets = netutil.bind_sockets(0, ip)
        server = httpserver.HTTPServer(app)
        server.add_sockets(sockets)
        for s in sockets:
            _logger.info('Listening on %s, port %d', *s.getsockname()[:2])
            port = s.getsockname()[1]
    if on_port is not None:
        on_port(port)
    return app, server


def main_server(on_port=None, closable=False, **params):
    app, server = init_app(on_port, closable, **params)
    io_loop = ioloop.IOLoop.current()
    if sys.platform.startswith('win'):
        # workaround for tornado on Windows:
        # add no-op to wake every 5s
        # to handle signals that may be ignored by the inner loop
        pc = ioloop.PeriodicCallback(lambda : None, 5000)
        pc.start()
    io_loop.start()
    # Clean up after server:
    server.stop()
    return app.exit_code


def _build_arg_parser():
    """
    Creates an argument parser that lets the user specify a port
    and displays a help message.
    """
    description = 'Web interface for Nbdime.'
    parser = ConfigBackedParser(description=description)
    add_generic_args(parser)
    add_web_args(parser)
    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    arguments = _build_arg_parser().parse_args(args)
    return main_server(port=arguments.port,
                       ip=arguments.ip,
                       cwd=arguments.workdirectory,
                       base_url=arguments.base_url,
                      )


if __name__ == '__main__':
    sys.exit(main())



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import pytest
import nbformat
from nbformat.v4 import new_notebook, new_code_cell
from collections import defaultdict

from nbdime import merge_notebooks, diff
from nbdime.diff_format import op_patch
from nbdime.utils import Strategies
from nbdime.diffing.config import DiffConfig
from nbdime.merging.generic import decide_merge, decide_merge_with_diff
from nbdime.merging.decisions import apply_decisions
from nbdime.merging.strategies import _cell_marker_format

from .utils import outputs_to_notebook, sources_to_notebook, strip_cell_ids, strip_cell_id


def test_decide_merge_strategy_fail(reset_log):
    """Check that "fail" strategy results in proper exception raised."""
    # One level dict
    base = {"foo": 1}
    local = {"foo": 2}
    remote = {"foo": 3}
    strategies = Strategies({"/foo": "fail"})
    with pytest.raises(RuntimeError):
        # pylint: disable=unused-variable
        conflicted_decisions = decide_merge(base, local, remote, strategies)

    # Nested dicts
    base = {"foo": {"bar": 1}}
    local = {"foo": {"bar": 2}}
    remote = {"foo": {"bar": 3}}
    strategies = Strategies({"/foo/bar": "fail"})
    with pytest.raises(RuntimeError):
        # pylint: disable=unused-variable
        decisions = decide_merge(base, local, remote, strategies)

    # We don't need this for non-leaf nodes and it's currently not implemented
    # strategies = Strategies({"/foo": "fail"})
    # with pytest.raises(RuntimeError):
    #     decisions = decide_merge(base, local, remote, strategies)


def test_decide_merge_strategy_clear1():
    """Check strategy "clear" in various cases."""
    # One level dict, clearing item value (think foo==execution_count)
    base = {"foo": 1}
    local = {"foo": 2}
    remote = {"foo": 3}
    strategies = Strategies({"/foo": "clear"})
    decisions = decide_merge(base, local, remote, strategies)
    assert apply_decisions(base, decisions) == {"foo": None}
    assert not any([d.conflict for d in decisions])

def test_decide_merge_strategy_clear2():
    base = {"foo": "1"}
    local = {"foo": "2"}
    remote = {"foo": "3"}
    strategies = Strategies({"/foo": "clear"})
    decisions = decide_merge(base, local, remote, strategies)
    #assert decisions == []
    assert apply_decisions(base, decisions) == {"foo": ""}
    assert not any([d.conflict for d in decisions])

    # We don't need this for non-leaf nodes and it's currently not implemented
    # base = {"foo": [1]}
    # local = {"foo": [2]}
    # remote = {"foo": [3]}
    # strategies = Strategies({"/foo": "clear"})
    # decisions = decide_merge(base, local, remote, strategies)
    # assert apply_decisions(base, decisions) == {"foo": []}
    # assert not any([d.conflict for d in decisions])


def test_decide_merge_strategy_clear_all():
    base = {"foo": [1, 2]}
    local = {"foo": [1, 4, 2]}
    remote = {"foo": [1, 3, 2]}

    strategies = Strategies({"/foo": "clear-all"})
    decisions = decide_merge(base, local, remote, strategies)
    assert apply_decisions(base, decisions) == {"foo": []}

    base = {"foo": [1, 2]}
    local = {"foo": [1, 4, 2]}
    remote = {"foo": [1, 2, 3]}

    strategies = Strategies({"/foo": "clear-all"})
    decisions = decide_merge(base, local, remote, strategies)
    assert apply_decisions(base, decisions) == {"foo": [1, 4, 2, 3]}


def test_decide_merge_strategy_remove():
    base = {"foo": [1, 2]}
    local = {"foo": [1, 4, 2]}
    remote = {"foo": [1, 3, 2]}

    strategies = Strategies({"/foo": "remove"})
    decisions = decide_merge(base, local, remote, strategies)
    assert apply_decisions(base, decisions) == {"foo": [1, 2]}
    assert decisions[0].local_diff != []
    assert decisions[0].remote_diff != []

    strategies = Strategies({})
    decisions = decide_merge(base, local, remote, strategies)
    assert apply_decisions(base, decisions) == {"foo": [1, 2]}
    assert decisions[0].local_diff != []
    assert decisions[0].remote_diff != []


def test_decide_merge_strategy_use_foo_on_dict_items():
    base = {"foo": 1}
    local = {"foo": 2}
    remote = {"foo": 3}

    strategies = Strategies({"/foo": "use-base"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": 1}

    strategies = Strategies({"/foo": "use-local"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": 2}

    strategies = Strategies({"/foo": "use-remote"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": 3}

    base = {"foo": {"bar": 1}}
    local = {"foo": {"bar": 2}}
    remote = {"foo": {"bar": 3}}

    strategies = Strategies({"/foo/bar": "use-base"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": {"bar": 1}}

    strategies = Strategies({"/foo/bar": "use-local"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": {"bar": 2}}

    strategies = Strategies({"/foo/bar": "use-remote"})
    decisions = decide_merge(base, local, remote, strategies)
    assert not any([d.conflict for d in decisions])
    assert apply_decisions(base, decisions) == {"foo": {"bar": 3}}


def test_decide_merge_simple_list_insert_conflict_resolution():
    # local and remote adds an entry each
    b = [1]
    l = [1, 2]
    r = [1, 3]

    strategies = Strategies({"/*": "use-local"})
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == l
    assert not any(d.conflict for d in decisions)

    strategies = Strategies({"/*": "use-remote"})
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == r
    assert not any(d.conflict for d in decisions)

    strategies = Strategies({"/*": "use-base"})
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == b
    assert not any(d.conflict for d in decisions)

    strategies = Strategies({"/": "clear-all"})
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == []
    assert not any(d.conflict for d in decisions)

def test_decide_merge_simple_list_insert_conflict_resolution__union():
    # local and remote adds an entry each
    b = [1]
    l = [1, 2]
    r = [1, 3]

    strategies = Strategies({"/": "union"})
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == [1, 2, 3]
    assert not any(d.conflict for d in decisions)


def test_decide_merge_list_conflicting_insertions_separate_chunks_v1():
    # local and remote adds an equal entry plus a different entry each
    # First, test when insertions DO NOT chunk together:
    b = [1, 9]
    l = [1, 2, 9, 11]
    r = [1, 3, 9, 11]

    # Check strategyless resolution
    strategies = Strategies({})
    resolved = decide_merge(b, l, r, strategies)
    expected_partial = [1, 9, 11]
    assert apply_decisions(b, resolved) == expected_partial
    assert len(resolved) == 2
    assert resolved[0].conflict
    assert not resolved[1].conflict

    strategies = Strategies({"/*": "use-local"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == l
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/*": "use-remote"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == r
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/*": "use-base"})
    resolved = decide_merge(b, l, r, strategies)
    # Strategy is only applied to conflicted decisions:
    assert apply_decisions(b, resolved) == expected_partial
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/": "clear-all"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == []
    assert not any(d.conflict for d in resolved)

    # from _merge_concurrent_inserts:
    # FIXME: This function doesn't work out so well with new conflict handling,
    # when an insert (e.g. [2,7] vs [3,7]) gets split into agreement on [7] and
    # conflict on [2] vs [3], the ordering gets lost. I think this was always
    # slightly ambiguous in the decision format, as the new inserts will get
    # the same key and decisions are supposed to be possible to reorder (sort)
    # without considering original ordering of decisions. To preserve the
    # ordering, perhaps we can add relative local/remote indices to the decisions?
    # We had this, where ordering made it work out correctly:
    #   "conflicting insert [2] vs [3] at 1 (base index);
    #    insert [7] at 1 (base index)"
    # instead we now have this which messes up the ordering:
    #   "insert [7] at 1 (base index);
    #    conflicting insert [2] vs [3] at 1 (base index)"
    # perhaps change to this:
    #   "insert [7] at key=1 (base index) lkey=1 rkey=1;
    #    conflicting insert [2] vs [3] at key=1 (base index) lkey=0 rkey=0"
    # then decisions can be sorted on (key,lkey) or (key,rkey) depending on chosen side.
    # This test covers the behaviour:
    # py.test -k test_shallow_merge_lists_insert_conflicted -s -vv
    #DEBUGGING = 1
    #if DEBUGGING: import ipdb; ipdb.set_trace()

    # Example:
    # b  l  r
    # 1  a  x
    # 2  b  y
    # 3  c  3
    # 4  4  4
    # Diffs:
    # b/l: insert a, b, c; remove 1-3
    # b/r: insert x, y; remove 1-2
    # The current chunking splits the removes here:
    # [insert a, b, c; remove 1-2]; [remove 3]
    # [insert x, y; remove 1-2]
    # That results in remove 3 not being conflicted.

def test_decide_merge_list_conflicting_insertions_separate_chunks_v2():
    # local and remote adds an equal entry plus a different entry each
    # First, test when insertions DO NOT chunk together:
    b = [1, 9]
    l = [1, 2, 9, 11]
    r = [1, 3, 9, 11]

    # Check strategyless resolution
    strategies = Strategies({})
    resolved = decide_merge(b, l, r, strategies)
    expected_partial = [1, 9, 11]
    assert apply_decisions(b, resolved) == expected_partial
    assert len(resolved) == 2
    assert resolved[0].conflict
    assert not resolved[1].conflict


def test_decide_merge_list_conflicting_insertions_separate_chunks__union():
    # local and remote adds an equal entry plus a different entry each
    # First, test when insertions DO NOT chunk together:
    b = [1, 9]
    l = [1, 2, 9, 11]
    r = [1, 3, 9, 11]

    strategies = Strategies({"/": "union"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == [1, 2, 3, 9, 11]
    assert not any(d.conflict for d in resolved)


def test_decide_merge_list_conflicting_insertions_in_chunks():
    # Next, test when insertions DO chunk together:
    b = [1, 9]
    l = [1, 2, 7, 9]
    r = [1, 3, 7, 9]

    # Check strategyless resolution
    strategies = Strategies({})
    resolved = decide_merge(b, l, r, strategies)
    expected_partial = [1, 7, 9]
    assert apply_decisions(b, resolved) == expected_partial

    strategies = Strategies({"/*": "use-local"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == l
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/*": "use-remote"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == r
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/*": "use-base"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == expected_partial
    assert not any(d.conflict for d in resolved)

    strategies = Strategies({"/": "clear-all"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == []
    assert not any(d.conflict for d in resolved)


def test_decide_merge_list_conflicting_insertions_in_chunks__union():
    # Next, test when insertions DO chunk together:
    b = [1, 9]
    l = [1, 2, 7, 9]
    r = [1, 3, 7, 9]

    strategies = Strategies({"/": "union"})
    resolved = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, resolved) == [1, 2, 3, 7, 9]
    assert not any(d.conflict for d in resolved)


def test_decide_merge_list_transients():
    # For this test, we need to use a custom predicate to ensure alignment
    common = {'id': 'This ensures alignment'}
    config = DiffConfig(
        predicates=defaultdict(lambda: [operator.__eq__], {
            '/': [lambda a, b: a['id'] == b['id']],
        })
    )

    # Setup transient difference in base and local, deletion in remote
    b = [{'transient': 22}]
    l = [{'transient': 242}]
    b[0].update(common)
    l[0].update(common)
    r = []

    # Make decisions based on diffs with predicates
    ld = diff(b, l, path="", config=config)
    rd = diff(b, r, path="", config=config)

    # Assert that generic merge without strategies gives conflict:
    strategies = Strategies()
    decisions = decide_merge_with_diff(b, l, r, ld, rd, strategies)
    assert len(decisions) == 1
    assert decisions[0].conflict
    assert apply_decisions(b, decisions) == b

    # Supply transient list to autoresolve, and check that transient is ignored
    strategies = Strategies(transients=[
        '/*/transient'
    ])
    decisions = decide_merge_with_diff(b, l, r, ld, rd, strategies)
    assert apply_decisions(b, decisions) == r
    assert not any(d.conflict for d in decisions)


def test_decide_merge_dict_transients():
    # Setup transient difference in base and local, deletion in remote
    b = {'a': {'transient': 22}}
    l = {'a': {'transient': 242}}
    r = {}

    # Assert that generic merge gives conflict
    strategies = Strategies()
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == b
    assert len(decisions) == 1
    assert decisions[0].conflict

    # Supply transient list to autoresolve, and check that transient is ignored
    strategies = Strategies(transients=[
        '/a/transient'
    ])
    decisions = decide_merge(b, l, r, strategies)
    assert apply_decisions(b, decisions) == r
    assert not any(d.conflict for d in decisions)


def test_decide_merge_mixed_nested_transients():
    # For this test, we need to use a custom predicate to ensure alignment
    common = {'id': 'This ensures alignment'}
    config = DiffConfig(
        predicates=defaultdict(lambda: [operator.__eq__], {
            '/': [lambda a, b: a['id'] == b['id']],
        })
    )
    # Setup transient difference in base and local, deletion in remote
    b = [{'a': {'transient': 22}}]
    l = [{'a': {'transient': 242}}]
    b[0].update(common)
    l[0].update(common)
    r = []

    # Make decisions based on diffs with predicates
    ld = diff(b, l, path="", config=config)
    rd = diff(b, r, path="", config=config)

    # Assert that generic merge gives conflict
    strategies = Strategies()
    decisions = decide_merge_with_diff(b, l, r, ld, rd, strategies)
    assert apply_decisions(b, decisions) == b
    assert len(decisions) == 1
    assert decisions[0].conflict

    # Supply transient list to autoresolve, and check that transient is ignored
    strategies = Strategies(transients=[
        '/*/a/transient'
    ])
    decisions = decide_merge_with_diff(b, l, r, ld, rd, strategies)
    assert apply_decisions(b, decisions) == r
    assert not any(d.conflict for d in decisions)


def test_inline_merge_empty_notebooks():
    "Missing fields all around passes through."
    base = {}
    local = {}
    remote = {}
    expected = {}
    merged, decisions = merge_notebooks(base, local, remote)
    assert expected == merged


def test_inline_merge_dummy_notebooks():
    "Just the basic empty notebook passes through."
    base = new_notebook()
    local = new_notebook()
    remote = new_notebook()
    expected = new_notebook()
    merged, decisions = merge_notebooks(base, local, remote)
    assert expected == merged


def test_inline_merge_notebook_version():
    "Minor version gets bumped to max."
    base = new_notebook(nbformat=4, nbformat_minor=0)
    local = new_notebook(nbformat=4, nbformat_minor=1)
    remote = new_notebook(nbformat=4, nbformat_minor=2)
    expected = new_notebook(nbformat=4, nbformat_minor=2)
    merged, decisions = merge_notebooks(base, local, remote)
    assert expected == merged


def test_inline_merge_notebook_metadata(reset_log):
    """Merging a wide range of different value types
    and conflict types in the root /metadata dicts.
    The goal is to exercise a decent part of the
    generic diff and merge functionality.
    """

    untouched = {
        "string": "untouched string",
        "integer": 123,
        "float": 16.0,
        "list": ["hello", "world"],
        "dict": {"first": "Hello", "second": "World"},
    }
    md_in = {
        1: {
            "untouched": untouched,
            "unconflicted": {
                "int_deleteme": 7,
                "string_deleteme": "deleteme",
                "list_deleteme": [7, "deleteme"],
                "dict_deleteme": {"deleteme": "now", "removeme": True},
                "list_deleteitem": [7, "deleteme", 3, "notme", 5, "deletemetoo"],

                "string": "string v1",
                "integer": 456,
                "float": 32.0,
                "list": ["hello", "universe"],
                "dict": {"first": "Hello", "second": "World", "third": "!"},
            },
            "conflicted": {
                "int_delete_replace": 3,
                "string_delete_replace": "string that will be deleted and modified",
                "list_delete_replace": [1],
                "dict_delete_replace": {"k":"v"},

            #     "string": "string v1",
            #     "integer": 456,
            #     "float": 32.0,
            #     "list": ["hello", "universe"],
            #     "dict": {"first": "Hello", "second": "World"},
            }
        },
        2: {
            "untouched": untouched,
            "unconflicted": {
                "dict_deleteme": {"deleteme": "now", "removeme": True},
                "list_deleteitem": [7, 3, "notme", 5, "deletemetoo"],

                "string": "string v1 equal addition",
                "integer": 123, # equal change
                "float": 16.0, # equal change
                # Equal delete at beginning and insert of two values at end:
                "list": ["universe", "new items", "same\non\nboth\nsides"],
                # cases covered: twosided equal value change, onesided delete, onesided replace, onesided insert, twosided insert of same value
                "dict": {"first": "changed", "second": "World", "third": "!", "newkey": "newvalue", "otherkey": "othervalue"},
            },
            "conflicted": {
                "int_delete_replace": 5,
                "list_delete_replace": [2],

                # "string": "another text",
                 #"integer": 456,
            #     "float": 16.0,
            #     "list": ["hello", "world"],
            #     "dict": {"new": "value", "first": "Hello"}, #"second": "World"},

            #     "added_string": "another text",
            #     "added_integer": 9,
            #     "added_float": 16.0,
            #     "added_list": ["another", "multiverse"],
            #     "added_dict": {"1st": "hey", "2nd": "there"},
            }
        },
        3: {
            "untouched": untouched,
            "unconflicted": {
                "list_deleteme": [7, "deleteme"],
                "list_deleteitem": [7, "deleteme", 3, "notme", 5],

                "string": "string v1 equal addition",
                "integer": 123, # equal change
                "float": 16.0, # equal change
                # Equal delete at beginning and insert of two values at end:
                "list": ["universe", "new items", "same\non\nboth\nsides"],
                "dict": {"first": "changed", "third": ".", "newkey": "newvalue"},
            },
            "conflicted": {
                "string_delete_replace": "string that is modified here and deleted in the other version",
                "dict_delete_replace": {"k":"x","q":"r"},

            #     "string": "different message",
            #     "integer": 456,
            #     #"float": 16.0,
            #     "list": ["hello", "again", "world"],
            #     "dict": {"new": "but different", "first": "Hello"}, #"second": "World"},

            #     "added_string": "but not the same string",
            #     #"added_integer": 9,
            #     "added_float": 64.0,
            #     "added_list": ["initial", "values", "another", "multiverse", "trailing", "values"],
            #     "added_dict": {"3rt": "mergeme", "2nd": "conflict"},
            }
        }
    }

    def join_dicts(dicta, dictb):
        d = {}
        d.update(dicta)
        d.update(dictb)
        return d

    shared_unconflicted = {
        "list_deleteitem": [7, 3, "notme", 5],

        "string": "string v1 equal addition",
        "integer": 123,
        "float": 16.0,
        "list": ["universe", "new items", "same\non\nboth\nsides"],
        "dict": {"first": "changed", "third": ".",  "newkey": "newvalue", "otherkey": "othervalue"},
    }
    shared_conflicted = {
        "int_delete_replace": 3,
        "string_delete_replace": "string that will be deleted and modified",
        "list_delete_replace": [1],
        "dict_delete_replace": {"k":"v"},

    #     #"string": "string v1",
    #     "string": "another textdifferent message",

    #     "float": 32.0,
    #     "list": ["hello", "universe"],
    #     "dict": {"first": "Hello", "second": "World"},
    #     # FIXME
    }

    md_out = {
        (1,2,3): {
            "untouched": untouched,
            "unconflicted": join_dicts(shared_unconflicted, {
                # ...
            }),
            "conflicted": join_dicts(shared_conflicted, {
                # ...
            }),
        },
        (1,3,2): {
            "untouched": untouched,
            "unconflicted": join_dicts(shared_unconflicted, {
                # ...
            }),
            "conflicted": join_dicts(shared_conflicted, {
                # ...
            }),
        },
    }

    # Fill in expected conflict records
    for triplet in sorted(md_out.keys()):
        i, j, k = triplet
        local_diff = diff(md_in[i]["conflicted"], md_in[j]["conflicted"])
        remote_diff = diff(md_in[i]["conflicted"], md_in[k]["conflicted"])

        # This may not be a necessary test, just checking my expectations
        assert local_diff == sorted(local_diff, key=lambda x: x.key)
        assert remote_diff == sorted(remote_diff, key=lambda x: x.key)

        c = {
            # These are patches on the /metadata dict
            "local_diff": [op_patch("conflicted", local_diff)],
            "remote_diff": [op_patch("conflicted", remote_diff)],
        }
        md_out[triplet]["nbdime-conflicts"] = c

    # Fill in the trivial merge results
    for i in (1, 2, 3):
        for j in (1, 2, 3):
            for k in (i, j):
                # For any combination i,j,i or i,j,j the
                # result should be j with no conflicts
                md_out[(i,j,k)] = md_in[j]

    tested = set()
    # Check the trivial merge results
    for i in (1, 2, 3):
        for j in (1, 2, 3):
            for k in (i, j):
                triplet = (i, j, k)
                tested.add(triplet)
                base = new_notebook(metadata=md_in[i])
                local = new_notebook(metadata=md_in[j])
                remote = new_notebook(metadata=md_in[k])
                # For any combination i,j,i or i,j,j the result should be j
                expected = new_notebook(metadata=md_in[j])
                merged, decisions = merge_notebooks(base, local, remote)
                assert "nbdime-conflicts" not in merged["metadata"]
                assert not any([d.conflict for d in decisions])
                assert expected == merged

    # Check handcrafted merge results
    for triplet in sorted(md_out.keys()):
        i, j, k = triplet
        tested.add(triplet)
        base = new_notebook(metadata=md_in[i])
        local = new_notebook(metadata=md_in[j])
        remote = new_notebook(metadata=md_in[k])
        expected = new_notebook(metadata=md_out[triplet])
        merged, decisions = merge_notebooks(base, local, remote)
        if "nbdime-conflicts" in merged["metadata"]:
            assert any([d.conflict for d in decisions])
        else:
            assert not any([d.conflict for d in decisions])
        assert expected == merged

    # At least try to run merge without crashing for permutations
    # of md_in that we haven't constructed expected results for
    for i in (1, 2, 3):
        for j in (1, 2, 3):
            for k in (1, 2, 3):
                triplet = (i, j, k)
                if triplet not in tested:
                    base = new_notebook(metadata=md_in[i])
                    local = new_notebook(metadata=md_in[j])
                    remote = new_notebook(metadata=md_in[k])
                    merged, decisions = merge_notebooks(base, local, remote)


def test_inline_merge_notebook_metadata_reproduce_bug(reset_log):
    md_in = {
        1: {
            "unconflicted": {
                "list_deleteitem": [7, "deleteme", 3, "notme", 5, "deletemetoo"],
            },
            "conflicted": {
                "dict_delete_replace": {"k":"v"},
            }
        },
        2: {
            "unconflicted": {
                "list_deleteitem": [7, 3, "notme", 5, "deletemetoo"],
            },
            "conflicted": {
            }
        },
        3: {
            "unconflicted": {
                "list_deleteitem": [7, "deleteme", 3, "notme", 5],
            },
            "conflicted": {
                "dict_delete_replace": {"k":"x"},
            }
        }
    }

    shared_unconflicted = {
        "list_deleteitem": [7, 3, "notme", 5],
    }
    shared_conflicted = {
        "dict_delete_replace": {"k":"v"},
    }

    md_out = {
        (1,2,3): {
            "unconflicted": shared_unconflicted,
            "conflicted": shared_conflicted
        },
    }

    # Fill in expected conflict records
    for triplet in sorted(md_out.keys()):
        i, j, k = triplet
        local_diff = diff(md_in[i]["conflicted"], md_in[j]["conflicted"])
        remote_diff = diff(md_in[i]["conflicted"], md_in[k]["conflicted"])

        # This may not be a necessary test, just checking my expectations
        assert local_diff == sorted(local_diff, key=lambda x: x.key)
        assert remote_diff == sorted(remote_diff, key=lambda x: x.key)

        c = {
            # These are patches on the /metadata dict
            "local_diff": [op_patch("conflicted", local_diff)],
            "remote_diff": [op_patch("conflicted", remote_diff)],
        }
        md_out[triplet]["nbdime-conflicts"] = c

    # Check handcrafted merge results
    triplet = (1,2,3)
    i, j, k = triplet
    base = new_notebook(metadata=md_in[i])
    local = new_notebook(metadata=md_in[j])
    remote = new_notebook(metadata=md_in[k])
    expected = new_notebook(metadata=md_out[triplet])
    merged, decisions = merge_notebooks(base, local, remote)
    if "nbdime-conflicts" in merged["metadata"]:
        assert any([d.conflict for d in decisions])
    else:
        assert not any([d.conflict for d in decisions])
    assert expected == merged


def test_inline_merge_source_empty():
    base = new_notebook()
    local = new_notebook()
    remote = new_notebook()
    expected = new_notebook()
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def code_nb(sources, strip_ids=False):
    nb = new_notebook(cells=[new_code_cell(s) for s in sources])
    strip_cell_ids(nb)
    return nb


def test_inline_merge_source_all_equal():
    base = code_nb([
        "first source",
        "other text",
        "yet more content",
    ])
    local = base
    remote = base
    expected = base
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_source_cell_deletions():
    "Cell deletions on both sides, onesided and agreed."
    base = code_nb([
        "first source",
        "other text",
        "yet more content",
        "and a final line",
        ])
    local = code_nb([
        #"first source",
        "other text",
        #"yet more content",
        #"and a final line",
        ])
    remote = code_nb([
        "first source",
        #"other text",
        "yet more content",
        #"and a final line",
        ])
    empty = code_nb([])
    for a in [base, local, remote, empty]:
        for b in [base, local, remote, empty]:
            merged, decisions = merge_notebooks(base, a, b)
            if a is b:
                assert merged == a
            elif a is base:
                assert merged == b
            elif b is base:
                assert merged == a
            else:
                # All other combinations will delete all cells
                assert merged == empty


def test_inline_merge_source_onesided_only():
    "A mix of changes on one side (delete, patch, remove)."
    base = code_nb([
        "first source",
        "other text",
        "yet more content",
        ])
    changed = code_nb([
        #"first source", # deleted
        "other text v2",
        "a different cell inserted",
        "yet more content",
        ])
    merged, decisions = merge_notebooks(base, changed, base)
    assert merged == changed
    merged, decisions = merge_notebooks(base, base, changed)
    assert merged == changed


def test_inline_merge_source_replace_line():
    "More elaborate test of cell deletions on both sides, onesided and agreed."
    # Note: Merge rendering of conflicted sources here will depend on git/diff/builtin params and availability
    base = code_nb([
        "first source",
        "other text",
        "this cell will be deleted and patched",
        "yet more content",
        "and a final line",
        ], strip_ids=True)
    local = code_nb([
        "1st source",  # onesided change
        "other text",
        #"this cell will be deleted and patched",
        "some more content",  # twosided equal change
        "And a Final line",  # twosided conflicted change
        ], strip_ids=True)
    remote = code_nb([
        "first source",
        "other text?",  # onesided change
        "this cell will be deleted and modified",
        "some more content",   # equal
        "and The final Line",  # conflicted
        ], strip_ids=True)
    expected = code_nb([
        "1st source",
        "other text?",
        #'<<<<<<< local <CELL DELETED>\n\n=======\nthis cell will be deleted and modified\n>>>>>>> remote'
        '<<<<<<< LOCAL CELL DELETED >>>>>>>\nthis cell will be deleted and modified',
        "some more content",  # equal
        '<<<<<<< local\nAnd a Final line\n=======\nand The final Line\n>>>>>>> remote'
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected
    expected = code_nb([
        "1st source",
        "other text?",
        #'<<<<<<< local\nthis cell will be deleted and modified\n=======\n>>>>>>> remote <CELL DELETED>'
        '<<<<<<< REMOTE CELL DELETED >>>>>>>\nthis cell will be deleted and modified',
        "some more content",
        '<<<<<<< local\nand The final Line\n=======\nAnd a Final line\n>>>>>>> remote'
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, remote, local)
    assert merged == expected


def test_inline_merge_source_add_to_line():
    "More elaborate test of cell deletions on both sides, onesided and agreed."
    # Note: Merge rendering of conflicted sources here will depend on git/diff/builtin params and availability
    base = code_nb([
        "first source",
        "other text",
        "this cell will be deleted and patched\nhere we add",
        "yet more content",
        "and a final line",
        ], strip_ids=True)
    local = code_nb([
        "1st source",  # onesided change
        "other text",
        #"this cell will be deleted and patched",
        "some more content",  # twosided equal change
        "And a Final line",  # twosided conflicted change
        ], strip_ids=True)
    remote = code_nb([
        "first source",
        "other text?",  # onesided change
        "this cell will be deleted and patched\nhere we add text to a line",
        "some more content",   # equal
        "and The final Line",  # conflicted
        ], strip_ids=True)
    expected = code_nb([
        "1st source",
        "other text?",
        #'<<<<<<< local <CELL DELETED>\n\n=======\nthis cell will be deleted and modified\n>>>>>>> remote'
        '<<<<<<< LOCAL CELL DELETED >>>>>>>\nthis cell will be deleted and patched\nhere we add text to a line',
        "some more content",  # equal
        '<<<<<<< local\nAnd a Final line\n=======\nand The final Line\n>>>>>>> remote'
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected
    expected = code_nb([
        "1st source",
        "other text?",
        #'<<<<<<< local\nthis cell will be deleted and modified\n=======\n>>>>>>> remote <CELL DELETED>'
        '<<<<<<< REMOTE CELL DELETED >>>>>>>\nthis cell will be deleted and patched\nhere we add text to a line',
        "some more content",
        '<<<<<<< local\nand The final Line\n=======\nAnd a Final line\n>>>>>>> remote'
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, remote, local)
    assert merged == expected


def test_inline_merge_source_patches_both_ends():
    "More elaborate test of cell deletions on both sides, onesided and agreed."
    # Note: Merge rendering of conflicted sources here will depend on git/diff/builtin params and availability
    base = code_nb([
        "first source will be modified",
        "other text",
        "this cell will be untouched",
        "yet more content",
        "and final line will be changed",
        ], strip_ids=True)
    local = code_nb([
        "first source will be modified locally",
        "other text",
        "this cell will be untouched",
        "yet more content",
        "and final line will be changed locally",
        ], strip_ids=True)
    remote = code_nb([
        "first source will be modified remotely",
        "other text",
        "this cell will be untouched",
        "yet more content",
        "and final line will be changed remotely",
        ], strip_ids=True)
    expected = code_nb([
        '<<<<<<< local\nfirst source will be modified locally\n=======\nfirst source will be modified remotely\n>>>>>>> remote',
        "other text",
        "this cell will be untouched",
        "yet more content",
        '<<<<<<< local\nand final line will be changed locally\n=======\nand final line will be changed remotely\n>>>>>>> remote',
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected
    expected = code_nb([
        '<<<<<<< local\nfirst source will be modified remotely\n=======\nfirst source will be modified locally\n>>>>>>> remote',
        "other text",
        "this cell will be untouched",
        "yet more content",
        '<<<<<<< local\nand final line will be changed remotely\n=======\nand final line will be changed locally\n>>>>>>> remote',
        ], strip_ids=True)
    merged, decisions = merge_notebooks(base, remote, local)
    assert merged == expected


def test_inline_merge_source_patch_delete_conflicts_both_ends():
    "More elaborate test of cell deletions on both sides, onesided and agreed."
    # Note: Merge rendering of conflicted sources here will depend on git/diff/builtin params and availability
    base = code_nb([
        "first source will be modified",
        "other text",
        "this cell will be untouched",
        "yet more content",
        "and final line will be changed",
        ])
    local = code_nb([
        "first source will be modified on one side",
        "other text",
        "this cell will be untouched",
        "yet more content",
        #"and final line will be deleted locally",
        ])
    remote = code_nb([
        #"first source will be deleted remotely",
        "other text",
        "this cell will be untouched",
        "yet more content",
        "and final line will be changed on one side",
        ])
    expected = code_nb([
        '<<<<<<< REMOTE CELL DELETED >>>>>>>\nfirst source will be modified on one side',
        "other text",
        "this cell will be untouched",
        "yet more content",
        '<<<<<<< LOCAL CELL DELETED >>>>>>>\nand final line will be changed on one side',
        ])
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected
    expected = code_nb([
        '<<<<<<< LOCAL CELL DELETED >>>>>>>\nfirst source will be modified on one side',
        "other text",
        "this cell will be untouched",
        "yet more content",
        '<<<<<<< REMOTE CELL DELETED >>>>>>>\nand final line will be changed on one side',
        ])
    merged, decisions = merge_notebooks(base, remote, local)
    assert merged == expected


def test_inline_merge_attachments():
    # FIXME: Use output creation utils Vidar wrote in another test file
    base = new_notebook()
    local = new_notebook()
    remote = new_notebook()
    expected = new_notebook()
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_outputs():
    # One cell with two outputs:
    base = outputs_to_notebook([['unmodified', 'base']], strip_ids=True)
    local = outputs_to_notebook([['unmodified', 'local']], strip_ids=True)
    remote = outputs_to_notebook([['unmodified', 'remote']], strip_ids=True)
    expected = outputs_to_notebook([[
        'unmodified',
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='<<<<<<< local <modified: text/plain>\n'),
        'local',
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='=======\n'),
        'remote',
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='>>>>>>> remote <modified: text/plain>\n'),
    ]], strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_outputs_conflicting_insert_in_empty():
    # One cell with two outputs:
    base = outputs_to_notebook([[]], strip_ids=True)
    local = outputs_to_notebook([['local']], strip_ids=True)
    remote = outputs_to_notebook([['remote']], strip_ids=True)
    expected = outputs_to_notebook([[
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='<<<<<<< local\n'),
        'local',
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='=======\n'),
        'remote',
        nbformat.v4.new_output(
            output_type='stream', name='stderr',
            text='>>>>>>> remote\n'),
    ]], strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_cells_insertion_similar():
    base = sources_to_notebook([['unmodified']], cell_type='markdown', strip_ids=True)
    local = sources_to_notebook([['unmodified'], ['local']], cell_type='markdown', strip_ids=True)
    remote = sources_to_notebook([['unmodified'], ['remote']], cell_type='markdown', strip_ids=True)
    expected = sources_to_notebook([
        'unmodified',
        [
            ("<"*7) + ' local\n',
            'local\n',
            ("="*7) + '\n',
            'remote\n',
            (">"*7) + ' remote'
        ]
    ], cell_type='markdown', strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_cells_insertion_unsimilar():
    base = sources_to_notebook([['unmodified']], cell_type='markdown', strip_ids=True)
    local = sources_to_notebook([['unmodified'], ['local\n', 'friendly faces\n', '3.14']], cell_type='markdown', strip_ids=True)
    remote = sources_to_notebook([['unmodified'], ['remote\n', 'foo bar baz\n']], cell_type='markdown', strip_ids=True)
    expected = sources_to_notebook([
        ['unmodified'],
        [_cell_marker_format(("<"*7) + ' local')],
        ['local\n', 'friendly faces\n', '3.14'],
        [_cell_marker_format("="*7)],
        ['remote\n', 'foo bar baz\n'],
        [_cell_marker_format((">"*7) + ' remote')],
    ], cell_type='markdown', strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert strip_cell_ids(merged) == expected


def test_inline_merge_cells_replacement_similar():
    base = sources_to_notebook([['unmodified'], ['base']], cell_type='markdown', strip_ids=False)
    local = sources_to_notebook([['unmodified'], ['local']], cell_type='markdown', strip_ids=True)
    remote = sources_to_notebook([['unmodified'], ['remote']], cell_type='markdown', strip_ids=True)
    expected = sources_to_notebook([
        ['unmodified'],
        [
            ("<"*7) + ' local\n',
            'local\n',
            ("="*7) + '\n',
            'remote\n',
            (">"*7) + ' remote'
        ]
    ], cell_type='markdown', strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert merged == expected


def test_inline_merge_cells_replacement_unsimilar():
    base = sources_to_notebook([['unmodified'], ['base']], cell_type='markdown', strip_ids=False)
    local = sources_to_notebook([['unmodified'], ['local\n', 'friendly faces\n', '3.14']], cell_type='markdown', strip_ids=True)
    remote = sources_to_notebook([['unmodified'], ['remote\n', 'foo bar baz\n']], cell_type='markdown', strip_ids=True)
    expected = sources_to_notebook([
        ['unmodified'],
        [_cell_marker_format(("<"*7) + ' local')],
        ['local\n', 'friendly faces\n', '3.14'],
        [_cell_marker_format("="*7)],
        ['remote\n', 'foo bar baz\n'],
        [_cell_marker_format((">"*7) + ' remote')],
    ], cell_type='markdown', strip_ids=True)
    merged, decisions = merge_notebooks(base, local, remote)
    assert strip_cell_ids(merged) == expected



# -*- coding: utf-8 -*-

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import json

import pytest
from tornado import ioloop
from tornado.httputil import url_concat
from tornado.escape import json_encode, json_decode
import nbformat

import nbdime.webapp.nbdiffweb
import nbdime.webapp.nbmergeweb

from .utils import WEB_TEST_TIMEOUT


diff_a = 'src-and-output--1.ipynb'
diff_b = 'src-and-output--2.ipynb'

merge_a = 'multilevel-test-base.ipynb'
merge_b = 'multilevel-test-local.ipynb'
merge_c = 'multilevel-test-remote.ipynb'



@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
def test_diff_web(filespath, unique_port, reset_log, ioloop_patch):
    a = os.path.join(filespath, diff_a)
    b = os.path.join(filespath, diff_b)
    loop = ioloop.IOLoop.current()
    loop.call_later(0, loop.stop)
    nbdime.webapp.nbdiffweb.main(['--port=%i' % unique_port, '--browser=disabled', a, b])


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
def test_diff_web_localhost(filespath, unique_port, reset_log, ioloop_patch):
    a = os.path.join(filespath, diff_a)
    b = os.path.join(filespath, diff_b)
    loop = ioloop.IOLoop.current()
    loop.call_later(0, loop.stop)
    nbdime.webapp.nbdiffweb.main([
        '--port=%i' % unique_port,
        '--browser=disabled',
        '--ip=localhost',
        a,
        b
    ])


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
def test_diff_web_gitrefs(git_repo2, unique_port, reset_log, ioloop_patch):
    a = 'local'
    b = 'remote'
    c = 'diff.ipynb'
    d = 'sub/subfile.ipynb'
    loop = ioloop.IOLoop.current()
    loop.call_later(0, loop.stop)
    nbdime.webapp.nbdiffweb.main(['--port=%i' % unique_port, '--browser=disabled', a, b, c, d])


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
def test_merge_web(filespath, unique_port, reset_log, ioloop_patch):
    a = os.path.join(filespath, merge_a)
    b = os.path.join(filespath, merge_b)
    c = os.path.join(filespath, merge_c)
    loop = ioloop.IOLoop.current()
    loop.call_later(0, loop.stop)
    nbdime.webapp.nbmergeweb.main(['--port=%i' % unique_port, '--browser=disabled', a, b, c])


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
@pytest.mark.gen_test
def test_fetch_diff(http_client, base_url, nbdime_base_url):
    url = url_concat(
        base_url + nbdime_base_url + '/diff',
        dict(base=diff_a, remote=diff_b))
    response = yield http_client.fetch(url)
    assert response.code == 200


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
@pytest.mark.gen_test
def test_api_diff(http_client, base_url, nbdime_base_url, diff_validator, filespath, auth_header):
    post_data = dict(base=diff_a, remote=diff_b)
    body = json_encode(post_data)

    url = base_url + nbdime_base_url + '/api/diff'
    response = yield http_client.fetch(url, method='POST', headers=auth_header, body=body)
    assert response.code == 200
    # Check that response is sane:
    data = json_decode(response.body)
    # Check that base is as expected:
    expected_base = nbformat.read(os.path.join(filespath, diff_a), as_version=4)
    assert json.dumps(data['base'], sort_keys=True) == json.dumps(expected_base, sort_keys=True)
    # Check that diff follows schema:
    diff_validator.validate(data['diff'])


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
@pytest.mark.gen_test
def test_fetch_merge(http_client, base_url, nbdime_base_url):
    url = url_concat(
        base_url + nbdime_base_url + '/merge',
        dict(base=merge_a, local=merge_b, remote=merge_c))
    response = yield http_client.fetch(url)
    assert response.code == 200


@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
@pytest.mark.gen_test
def test_api_merge(http_client, base_url, nbdime_base_url, merge_validator, filespath, auth_header):
    post_data = dict(base=merge_a, local=merge_b, remote=merge_c)
    body = json_encode(post_data)

    url = base_url + nbdime_base_url + '/api/merge'
    response = yield http_client.fetch(url, method='POST', headers=auth_header, body=body)
    assert response.code == 200
    # Check that response is sane:
    data = json_decode(response.body)
    # Check that base is as expected:
    expected_base = nbformat.read(os.path.join(filespath, merge_a), as_version=4)
    assert json.dumps(data['base'], sort_keys=True) == json.dumps(expected_base, sort_keys=True)
    # Check that decisions follows schema:
    merge_validator.validate(data['merge_decisions'])




"""Tools for profiling diff/merge performance.

Sometimes diffing times can become unacceptably long.
Then the tools in this module can help you profile and figure out
where the time is spent. For a more rigorous profiling, consider using
cPython, but for initial considerations these should be helpful.

Typical profiling usage:
Add some statements like
from nbdime.profiling import timer
with timer.time('Key to identify this segment'):
    <code to time>

Then, launch `python -m nbdime.profiling notebookA.ipynb notebookB.ipynb`.

Example case:
Output diffing is slow. By cPython/timer, the time spent has been narrowed
to comparison of MIME types (`_compare_mimedata()`). To figure out which
MIME type is taking all the time, the following is added to that function:

    with timer.time(mimetype):
        <existing function body>

The output becomes:

    Key           Calls          Time
    ----------  -------  ------------
    text/html         6  74.9686
    text/plain       12   0.000500202
    image/png         5   0

From this we can see that the HTML comparison is really blowing up!
Further inspection reveals that there are two html outputs in each
notebook that contain an embedded base64 image. These caused the text
diff to blow up.

"""

import time
import contextlib
from tabulate import tabulate
from functools import wraps


def _sort_time(value):
    time = value[1]['time']
    return -time


class TimePaths(object):
    def __init__(self, verbose=False, enabled=True):
        self.verbose = verbose
        self.map = {}
        self.enabled = enabled

    @contextlib.contextmanager
    def time(self, key):
        if not self.enabled:
            yield
            return
        start = time.time()
        yield
        end = time.time()
        secs = end - start
        if key in self.map:
            self.map[key]['time'] += secs
            self.map[key]['calls'] += 1
        else:
            self.map[key] = dict(time=secs, calls=1)

    def profile(self, key=None):
        def decorator(function):
            nonlocal key
            if key is None:
                key = function.__name__ or 'unknown'
            @wraps(function)
            def inner(*args, **kwargs):
                with self.time(key):
                    return function(*args, **kwargs)
            return inner
        return decorator

    @contextlib.contextmanager
    def enable(self):
        old = self.enabled
        self.enabled = True
        yield
        self.enabled = old

    @contextlib.contextmanager
    def disable(self):
        old = self.enabled
        self.enabled = False
        yield
        self.enabled = old

    def __str__(self):
        # First, sort by path
        items = sorted(self.map.items(), key=_sort_time)
        lines = []
        for key, data in items:
            time = data['time']
            calls = data['calls']
            lines.append((key, calls, time, time / calls))
        return tabulate(lines, headers=['Key', 'Calls', 'Time', 'Time/Call'])


timer = TimePaths(enabled=False)


def profile_diff_paths(args=None):
    import nbdime.nbdiffapp
    import nbdime.profiling
    try:
        with nbdime.profiling.timer.enable():
            nbdime.nbdiffapp.main(args)
    finally:
        data = str(nbdime.profiling.timer)
        print(data)


if __name__ == "__main__":
    profile_diff_paths()




import argparse
import pytest
import json

from traitlets import Enum

from nbdime.args import (
    SkipAction, ConfigBackedParser, LogLevelAction,
)
from nbdime.config import (
    entrypoint_configurables, Global, _Ignorables
)
import nbdime.diffing.notebooks
from nbdime.diffing.notebooks import notebook_differs, diff


class FixtureConfig(Global):
    log_level = Enum(
        ('DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'),
        'WARN',
    ).tag(config=True)

@pytest.fixture
def entrypoint_config():
    entrypoint_configurables['test-prog'] = FixtureConfig
    yield
    del entrypoint_configurables['test-prog']

class IgnorableConfig1(_Ignorables):
    pass

class IgnorableConfig2(IgnorableConfig1):
    pass

@pytest.fixture
def entrypoint_ignore_config():
    entrypoint_configurables['test-prog'] = IgnorableConfig2
    yield
    del entrypoint_configurables['test-prog']


def test_skip_action():
    parser = argparse.ArgumentParser()
    parser.add_argument('a')
    parser.add_argument('b', action=SkipAction)

    opts = parser.parse_args(['fileA', 'fileB'])

    assert opts.a == 'fileA'
    assert not hasattr(opts, 'b')


def test_config_parser(entrypoint_config):
    parser = ConfigBackedParser('test-prog')
    parser.add_argument(
        '--log-level',
        default='INFO',
        choices=('DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'),
        help="Set the log level by name.",
        action=LogLevelAction,
    )

    # Check that log level default is taken from FixtureConfig
    arguments = parser.parse_args([])
    assert arguments.log_level == 'WARN'

    arguments = parser.parse_args(['--log-level', 'ERROR'])
    assert arguments.log_level == 'ERROR'


def test_ignore_config_simple(entrypoint_ignore_config, tmpdir, reset_notebook_diff):
    tmpdir.join('nbdime_config.json').write_text(
        json.dumps({
            'IgnorableConfig1': {
                'Ignore': {
                    '/cells/*/metadata': ['collapsed', 'autoscroll']
                }
            },
        }),
        encoding='utf-8'
    )

    def mock_ignore_keys(inner, keys):
        return (inner, keys)
    parser = ConfigBackedParser('test-prog')
    old_keys = nbdime.diffing.notebooks.diff_ignore_keys
    nbdime.diffing.notebooks.diff_ignore_keys = mock_ignore_keys
    try:
        with tmpdir.as_cwd():
            parser.parse_args([])
    finally:
        nbdime.diffing.notebooks.diff_ignore_keys = old_keys

    assert notebook_differs['/cells/*/metadata'] == (diff, ['collapsed', 'autoscroll'])



def test_ignore_config_merge(entrypoint_ignore_config, tmpdir, reset_notebook_diff):
    tmpdir.join('nbdime_config.json').write_text(
        json.dumps({
            'IgnorableConfig1': {
                'Ignore': {
                    '/cells/*/metadata': ['collapsed', 'autoscroll']
                }
            },
            'IgnorableConfig2': {
                'Ignore': {
                    '/metadata': ['foo'],
                    '/cells/*/metadata': ['tags']
                }
            },
        }),
        encoding='utf-8'
    )

    def mock_ignore_keys(inner, keys):
        return (inner, keys)
    parser = ConfigBackedParser('test-prog')
    old_keys = nbdime.diffing.notebooks.diff_ignore_keys
    nbdime.diffing.notebooks.diff_ignore_keys = mock_ignore_keys
    try:
        with tmpdir.as_cwd():
            parser.parse_args([])
    finally:
        nbdime.diffing.notebooks.diff_ignore_keys = old_keys

    assert notebook_differs['/metadata'] == (diff, ['foo'])
    # Lists are not merged:
    assert notebook_differs['/cells/*/metadata'] == (diff, ['tags'])


def test_config_inherit(entrypoint_ignore_config, tmpdir, reset_notebook_diff):
    tmpdir.join('nbdime_config.json').write_text(
        json.dumps({
            'IgnorableConfig1': {
                'metadata': False
            },
        }),
        encoding='utf-8'
    )

    parser = ConfigBackedParser('test-prog')
    with tmpdir.as_cwd():
        parsed = parser.parse_args([])

    assert parsed.metadata is False



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import operator

__all__ = ["diff_sequence_myers"]


def diff_sequence_myers(A, B, compare=operator.__eq__):
    """Compute the diff of A and B using Myers' O(ND) algorithm."""
    raise NotImplementedError("Work in progress...")



#!/usr/bin/env python
"""A mercurial CLI merge tool for notebooks.

Uses nbdime to merge notebooks instead of plain text merges of JSON.
See the documentation for how to correctly configure mercurial to use this.

Use with:

    hg merge [<commit> [<commit>]]
"""



import argparse
import sys

from nbdime import nbmergeapp
from nbdime.args import (
    add_generic_args, add_diff_args, add_merge_args, add_filename_args,
    ConfigBackedParser,
)



def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = ConfigBackedParser('hg-nbmerge', description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    add_generic_args(parser)
    add_diff_args(parser)
    add_merge_args(parser)

    # Argument list, we are given base, local, remote
    add_filename_args(parser, ["base", "local", "remote", "merged"])

    opts = parser.parse_args(args)
    # mergeapp expects an additional decisions arg:
    opts.decisions = False
    opts.out = opts.merged
    del opts.merged
    return nbmergeapp.main_merge(opts)


if __name__ == "__main__":
    sys.exit(main())




class DiffConfig:
    """Set of predicates/differs/other configs to pass around"""

    def __init__(self, *, predicates=None, differs=None, atomic_paths=None):
        if predicates is None:
            from .generic import default_predicates
            predicates = default_predicates()

        if differs is None:
            from .generic import default_differs
            differs = default_differs()

        self.predicates = predicates
        self.differs = differs
        self._atomic_paths = atomic_paths or {}

    def diff_item_at_path(self, a, b, path):
        """Calculate the diff for path."""
        self.differs[path](a, b, path=path, config=self)

    def is_atomic(self, x, path=None):
        "Return True for values that diff should treat as a single atomic value."
        try:
            return self._atomic_paths[path]
        except KeyError:
            return not isinstance(x, (str, list, dict))

    def __copy__(self):
        return DiffConfig(
            predicates=self.predicates.copy(),
            differs=self.differs.copy(),
            atomic_paths=self._atomic_paths.copy(),
        )
